from __future__ import annotations
import abc
import hashlib
from typing import TypeVar, Type
from neo3 import vm
from enum import IntEnum
from events import Events  # type: ignore
from neo3.core import types, jsonpath


msgrouter = Events()

# :noindex:


class Size(IntEnum):
    """
    Explicit bytes of memory consumed
    """
    uint8 = 1
    uint16 = 2
    uint32 = 4
    uint64 = 8
    uint160 = 20
    uint256 = 32


class IClonable(abc.ABC):
    @abc.abstractmethod
    def clone(self):
        """
        Create a deep copy of self
        """

    def from_replica(self, replica):
        """ shallow copy from a replica object """


class IJson(abc.ABC):
    @abc.abstractmethod
    def to_json(self) -> dict:
        """ convert object into json """

    @classmethod
    @abc.abstractmethod
    def from_json(cls, json: dict):
        """ create object from JSON """


T = TypeVar('T', bound='IInteroperable')


class IInteroperable(abc.ABC):
    @abc.abstractmethod
    def to_stack_item(self, reference_counter: vm.ReferenceCounter) -> vm.StackItem:
        """ Convert object to a virtual machine stack item"""

    @classmethod
    def from_stack_item(cls: Type[T], stack_item: vm.StackItem) -> T:
        """ Convert a stack item into an object"""
        raise ValueError(f"{cls.__class__.__name__} cannot be converted to a stack item")


def to_script_hash(data: bytes) -> types.UInt160:
    """
    Create a script hash based on the input data.

    Args:
        data: data to hash
    """
    intermediate_data = hashlib.sha256(data).digest()
    data_ = hashlib.new('ripemd160', intermediate_data).digest()
    return types.UInt160(data_)


def syscall_name_to_int(name: str) -> int:
    return int.from_bytes(hashlib.sha256(name.encode()).digest()[:4], 'little', signed=False)



from enum import IntFlag


class CallFlags(IntFlag):
    """
    Describes the required call permissions for contract functions.
    """
    NONE = 0,
    READ_STATES = 0x1
    WRITE_STATES = 0x02
    ALLOW_CALL = 0x04
    ALLOW_NOTIFY = 0x08
    STATES = READ_STATES | WRITE_STATES
    READ_ONLY = READ_STATES | ALLOW_CALL
    ALL = STATES | ALLOW_CALL | ALLOW_NOTIFY

    @classmethod
    def from_csharp_name(cls, input: str):
        if input == "None":
            return CallFlags.NONE
        elif input == "ReadStates":
            return CallFlags.READ_STATES
        elif input == "WriteStates":
            return CallFlags.WRITE_STATES
        elif input == "AllowCall":
            return CallFlags.ALLOW_CALL
        elif input == "AllowNotify":
            return CallFlags.ALLOW_NOTIFY
        elif input == "States":
            return CallFlags.STATES
        elif input == "ReadOnly":
            return CallFlags.READ_ONLY
        elif input == "All":
            return CallFlags.ALL
        else:
            raise ValueError(f"{input} is not a valid member of {cls.__name__}")



from __future__ import annotations

from jsonschema import validate  # type: ignore

from neo3.core import IJson


class ScryptParameters(IJson):

    json_schema = {
        "type": "object",
        "properties": {
            "n": {"type": "integer"},
            "r": {"type": "integer"},
            "p": {"type": "integer"}
        },
        "required": ["n", "r", "p"]
    }

    def __init__(self,
                 n: int = 16384,
                 r: int = 8,
                 p: int = 8):

        self.n = n
        self.r = r
        self.p = p

    def to_json(self) -> dict:
        return {
            'n': self.n,
            'r': self.r,
            'p': self.p
        }

    @classmethod
    def from_json(cls, json: dict) -> ScryptParameters:
        """
        Parse object out of JSON data.

        Args:
            json: a dictionary.

        Raises:
            KeyError: if the data supplied does not contain the necessary key.
        """
        validate(json, schema=cls.json_schema)

        return cls(n=json['n'],
                   r=json['r'],
                   p=json['p'])



"""The actual classes come from C-extension modules"""
from __future__ import annotations
import hashlib
from pybiginteger import BigInteger
from neo3vm import ScriptBuilder as _ScriptBuilder
from neo3vm import *

del globals()['ScriptBuilder']
from neo3vm import ScriptBuilder as _ScriptBuilder


def _syscall_name_to_int(name: str) -> int:
    return int.from_bytes(hashlib.sha256(name.encode()).digest()[:4], 'little', signed=False)


class ScriptBuilder(_ScriptBuilder):  # type: ignore
    def emit_dynamic_call(self, script_hash, operation: str) -> None:
        self.emit(OpCode.NEWARRAY0)
        self.emit_push(0xF)  # CallFlags.ALL
        self.emit_push(operation)
        self.emit_push(script_hash.to_array())
        self.emit_syscall(_syscall_name_to_int("System.Contract.Call"))

    def emit_dynamic_call_with_args(self, script_hash, operation: str, args) -> None:
        for arg in reversed(args):
            self.emit_push(arg)
        self.emit_push(len(args))
        self.emit(OpCode.PACK)
        self.emit_push(0xF)  # CallFlags.ALL
        self.emit_push(operation)
        self.emit_push(script_hash.to_array())
        self.emit_syscall(_syscall_name_to_int("System.Contract.Call"))



from __future__ import annotations
import struct
from .nativecontract import NativeContract
from .decorator import register
from neo3 import storage, contracts, vm, settings
from neo3.core import types, msgrouter, cryptography, serialization, to_script_hash, Size as s, IInteroperable
from typing import Tuple, List, Dict, Sequence, cast, Optional


class FungibleTokenStorageState(IInteroperable, serialization.ISerializable):
    """
    Helper class for NEP17 balance state
    """

    def __init__(self):
        super(FungibleTokenStorageState, self).__init__()
        self.balance: vm.BigInteger = vm.BigInteger.zero()

    def __len__(self):
        return len(self.balance.to_array())

    def serialize(self, writer: serialization.BinaryWriter) -> None:
        writer.write_var_bytes(self.balance.to_array())

    def deserialize(self, reader: serialization.BinaryReader) -> None:
        self.balance = vm.BigInteger(reader.read_var_bytes())

    def to_stack_item(self, reference_counter: vm.ReferenceCounter) -> vm.StackItem:
        struct_ = vm.StructStackItem(reference_counter)
        struct_.append(vm.IntegerStackItem(self.balance))
        return struct_

    @classmethod
    def from_stack_item(cls, stack_item: vm.StackItem):
        si = cast(vm.StructStackItem, stack_item)
        c = cls()
        c.balance = si[0].to_biginteger()
        return c


class FungibleToken(NativeContract):
    _id: int = -99999
    _decimals: int = -1

    _state = FungibleTokenStorageState
    _symbol: str = ""

    def init(self):
        super(FungibleToken, self).init()

        self.key_account = storage.StorageKey(self._id, b'\x14')
        self.key_total_supply = storage.StorageKey(self._id, b'\x0B')

        self.manifest.supported_standards = ["NEP-17"]
        self.manifest.abi.events = [
            contracts.ContractEventDescriptor(
                "Transfer",
                parameters=[
                    contracts.ContractParameterDefinition("from", contracts.ContractParameterType.HASH160),
                    contracts.ContractParameterDefinition("to", contracts.ContractParameterType.HASH160),
                    contracts.ContractParameterDefinition("amount", contracts.ContractParameterType.INTEGER)
                ]
            )
        ]
        self.factor = pow(vm.BigInteger(10), vm.BigInteger(self._decimals))

    @register("decimals", contracts.CallFlags.READ_STATES)
    def decimals(self) -> int:
        return self._decimals

    @register("symbol", contracts.CallFlags.READ_STATES)
    def symbol(self) -> str:
        """ Token symbol. """
        return self._symbol

    @register("totalSupply", contracts.CallFlags.READ_STATES, cpu_price=1 << 15)
    def total_supply(self, snapshot: storage.Snapshot) -> vm.BigInteger:
        """ Get the total deployed tokens. """
        storage_item = snapshot.storages.try_get(self.key_total_supply)
        if storage_item is None:
            return vm.BigInteger.zero()
        else:
            return vm.BigInteger(storage_item.value)

    @register("balanceOf", contracts.CallFlags.READ_STATES, cpu_price=1 << 15)
    def balance_of(self, snapshot: storage.Snapshot, account: types.UInt160) -> vm.BigInteger:
        """
        Get the balance of an account.

        Args:
            snapshot: snapshot of the storage
            account: script hash of the account to obtain the balance of

        Returns:
            amount of balance.

        Note: The returned value is still in internal format. Divide the results by the contract's `decimals`
        """
        storage_item = snapshot.storages.try_get(self.key_account + account, read_only=True)
        if storage_item is None:
            return vm.BigInteger.zero()
        else:
            state = self._state.deserialize_from_bytes(storage_item.value)
            return state.balance

    @register("transfer",
              (contracts.CallFlags.STATES | contracts.CallFlags.ALLOW_CALL | contracts.CallFlags.ALLOW_NOTIFY),
              cpu_price=1 << 17, storage_price=50)
    def transfer(self,
                 engine: contracts.ApplicationEngine,
                 account_from: types.UInt160,
                 account_to: types.UInt160,
                 amount: vm.BigInteger,
                 data: vm.StackItem
                 ) -> bool:
        """
        Transfer tokens from one account to another.

        Raises:
            ValueError: if the requested amount is negative.

        Returns:
            True on success. False otherwise.
        """
        if amount.sign < 0:
            raise ValueError("Can't transfer a negative amount")

        # transfer from an account not owned by the smart contract that is requesting the transfer
        # and there is no signature that approves we are allowed todo so
        if account_from != engine.calling_scripthash and not engine.checkwitness(account_from):
            return False

        storage_key_from = self.key_account + account_from
        storage_item_from = engine.snapshot.storages.try_get(storage_key_from, read_only=False)

        if amount == vm.BigInteger.zero():
            if storage_item_from is not None:
                state_from = storage_item_from.get(self._state)
                self.on_balance_changing(engine, account_from, state_from, amount)
        else:
            if storage_item_from is None:
                return False

            state_from = storage_item_from.get(self._state)
            if state_from.balance < amount:
                return False

            if account_from == account_to:
                self.on_balance_changing(engine, account_from, state_from, vm.BigInteger.zero())
            else:
                self.on_balance_changing(engine, account_from, state_from, -amount)
                if state_from.balance == amount:
                    engine.snapshot.storages.delete(storage_key_from)
                else:
                    state_from.balance -= amount

                storage_key_to = self.key_account + account_to
                storage_item_to = engine.snapshot.storages.try_get(storage_key_to, read_only=False)
                if storage_item_to is None:
                    storage_item_to = storage.StorageItem(self._state().to_array())
                    engine.snapshot.storages.put(storage_key_to, storage_item_to)

                state_to = storage_item_to.get(self._state)

                self.on_balance_changing(engine, account_to, state_to, amount)
                state_to.balance += amount

        self._post_transfer(engine, account_from, account_to, amount, data, True)
        return True

    def mint(self,
             engine: contracts.ApplicationEngine,
             account: types.UInt160,
             amount: vm.BigInteger,
             call_on_payment: bool) -> None:
        """
        Mint an amount of tokens into account.

        Increases the total supply of the token.
        """
        if amount.sign < 0:
            raise ValueError("Can't mint a negative amount")

        if amount == vm.BigInteger.zero():
            return

        storage_key = self.key_account + account
        storage_item = engine.snapshot.storages.try_get(storage_key, read_only=False)

        if storage_item is None:
            storage_item = storage.StorageItem(self._state().to_array())
            engine.snapshot.storages.put(storage_key, storage_item)

        state = storage_item.get(self._state)
        self.on_balance_changing(engine, account, state, amount)
        state.balance += amount

        storage_item = engine.snapshot.storages.try_get(self.key_total_supply, read_only=False)
        if storage_item is None:
            storage_item = storage.StorageItem(amount.to_array())
            engine.snapshot.storages.put(self.key_total_supply, storage_item)
        else:
            old_value = vm.BigInteger(storage_item.value)
            storage_item.value = (amount + old_value).to_array()
        self._post_transfer(engine, types.UInt160.zero(), account, amount, vm.NullStackItem(), call_on_payment)

    def burn(self, engine: contracts.ApplicationEngine, account: types.UInt160, amount: vm.BigInteger) -> None:
        """
        Burn an amount of tokens from account.

        Reduces the total supply of the token.

        Raises:
            ValueError: if amount is negative
            ValueError: if the burn amount is larger than the available balance in the account
        """
        if amount.sign < 0:
            raise ValueError("Can't burn a negative amount")

        if amount == vm.BigInteger.zero():
            return

        storage_key = self.key_account + account
        storage_item = engine.snapshot.storages.get(storage_key, read_only=False)

        state = storage_item.get(self._state)
        if state.balance < amount:
            raise ValueError(f"Insufficient balance. Requesting to burn {amount}, available {state.balance}")

        self.on_balance_changing(engine, account, state, -amount)
        if state.balance == amount:
            engine.snapshot.storages.delete(storage_key)
        else:
            state.balance -= amount
            engine.snapshot.storages.update(storage_key, storage_item)

        storage_item = engine.snapshot.storages.get(self.key_total_supply, read_only=False)
        old_value = vm.BigInteger(storage_item.value)
        new_value = old_value - amount
        if new_value == vm.BigInteger.zero():
            engine.snapshot.storages.delete(self.key_total_supply)
        else:
            storage_item.value = new_value.to_array()
        self._post_transfer(engine, account, types.UInt160.zero(), amount, vm.NullStackItem(), False)

    def on_balance_changing(self, engine: contracts.ApplicationEngine,
                            account: types.UInt160,
                            state,
                            amount: vm.BigInteger) -> None:
        pass

    def on_persist(self, engine: contracts.ApplicationEngine) -> None:
        pass

    def _post_transfer(self,
                       engine: contracts.ApplicationEngine,
                       account_from: types.UInt160,
                       account_to: types.UInt160,
                       amount: vm.BigInteger,
                       data: vm.StackItem,
                       call_on_payment: bool) -> None:
        state = vm.ArrayStackItem(vm.ReferenceCounter())
        if account_from == types.UInt160.zero():
            state.append(vm.NullStackItem())
        else:
            state.append(vm.ByteStringStackItem(account_from.to_array()))
        if account_to == types.UInt160.zero():
            state.append(vm.NullStackItem())
        else:
            state.append(vm.ByteStringStackItem(account_to.to_array()))
        state.append(vm.IntegerStackItem(amount))

        msgrouter.interop_notify(self.hash, "Transfer", state)

        # wallet or smart contract
        if not call_on_payment \
                or account_to == types.UInt160.zero() \
                or contracts.ManagementContract().get_contract(engine.snapshot, account_to) is None:
            return

        if account_from == types.UInt160.zero():
            from_: vm.StackItem = vm.NullStackItem()
        else:
            from_ = vm.ByteStringStackItem(account_from.to_array())
        engine.call_from_native(self.hash, account_to, "onNEP17Payment", [from_, vm.IntegerStackItem(amount), data])


class _NeoTokenStorageState(FungibleTokenStorageState):
    """
    Helper class for storing voting and bonus GAS state
    """

    def __init__(self):
        super(_NeoTokenStorageState, self).__init__()
        self.vote_to: cryptography.ECPoint = cryptography.ECPoint.deserialize_from_bytes(b'\x00')
        self.balance_height: int = 0

    def __len__(self):
        return super(_NeoTokenStorageState, self).__len__() + len(self.vote_to) + s.uint32

    def serialize(self, writer: serialization.BinaryWriter) -> None:
        super(_NeoTokenStorageState, self).serialize(writer)
        writer.write_serializable(self.vote_to)
        writer.write_uint32(self.balance_height)

    def deserialize(self, reader: serialization.BinaryReader) -> None:
        super(_NeoTokenStorageState, self).deserialize(reader)
        self.vote_to = reader.read_serializable(cryptography.ECPoint)  # type: ignore
        self.balance_height = reader.read_uint32()

    def to_stack_item(self, reference_counter: vm.ReferenceCounter) -> vm.StackItem:
        struct_ = super(_NeoTokenStorageState, self).to_stack_item(reference_counter)
        struct_ = cast(vm.StructStackItem, struct_)
        struct_.append(vm.IntegerStackItem(self.balance_height))
        if self.vote_to.is_zero():
            struct_.append(vm.NullStackItem())
        else:
            struct_.append(vm.ByteStringStackItem(self.vote_to.to_array()))
        return struct_

    @classmethod
    def from_stack_item(cls, stack_item: vm.StackItem):
        c = super(_NeoTokenStorageState, cls).from_stack_item(stack_item)  # type: _NeoTokenStorageState
        si = cast(vm.StructStackItem, stack_item)
        c.balance_height = int(si[1].to_biginteger())
        c.vote_to = cryptography.ECPoint.deserialize_from_bytes(si[2].to_array())
        return c


class _CandidateState(serialization.ISerializable):
    """
    Helper class for storing consensus candidates and their votes
    """

    def __init__(self):
        self.registered = True
        self.votes = vm.BigInteger.zero()

    def __len__(self):
        return 1 + len(self.votes.to_array())

    def serialize(self, writer: serialization.BinaryWriter) -> None:
        writer.write_bool(self.registered)
        writer.write_var_bytes(self.votes.to_array())

    def deserialize(self, reader: serialization.BinaryReader) -> None:
        self.registered = reader.read_bool()
        self.votes = vm.BigInteger(reader.read_var_bytes())


class _CommitteeState(serialization.ISerializable):
    def __init__(self,
                 snapshot: storage.Snapshot,
                 validators: Dict[cryptography.ECPoint, vm.BigInteger]):
        self.snapshot = snapshot
        self._validators = validators
        self._storage_key = NeoToken().key_committee

    def __len__(self):
        return len(self.to_array())

    def __getitem__(self, item: cryptography.ECPoint) -> vm.BigInteger:
        return self._validators[item]

    @classmethod
    def from_snapshot(cls, snapshot: storage.Snapshot):
        c = cls(snapshot, {})
        with serialization.BinaryReader(snapshot.storages.get(c._storage_key, read_only=True).value) as reader:
            c.deserialize(reader)
        return c

    @property
    def validators(self) -> List[cryptography.ECPoint]:
        return list(self._validators.keys())

    def persist(self, snapshot: storage.Snapshot):
        self.snapshot = snapshot
        self.snapshot.storages.update(self._storage_key, storage.StorageItem(self.to_array()))

    def serialize(self, writer: serialization.BinaryWriter) -> None:
        writer.write_var_int(len(self._validators))
        for key, value in self._validators.items():
            writer.write_serializable(key)
            writer.write_var_bytes(value.to_array())

    def deserialize(self, reader: serialization.BinaryReader) -> None:
        length = reader.read_var_int()
        if self._validators is None:
            self._validators = {}
        else:
            self._validators.clear()

        for _ in range(length):
            public_key = reader.read_serializable(cryptography.ECPoint)  # type: ignore
            self._validators.update({
                public_key: vm.BigInteger(reader.read_var_bytes())
            })


class _GasRecord(serialization.ISerializable):
    def __init__(self, index: int, gas_per_block: vm.BigInteger):
        self.index = index
        self.gas_per_block = gas_per_block

    def __len__(self):
        return s.uint32 + len(self.gas_per_block.to_array())

    def serialize(self, writer: serialization.BinaryWriter) -> None:
        writer.write_uint32(self.index)
        writer.write_var_bytes(self.gas_per_block.to_array())

    def deserialize(self, reader: serialization.BinaryReader) -> None:
        self.index = reader.read_uint32()
        self.gas_per_block = vm.BigInteger(reader.read_var_bytes())

    @classmethod
    def _serializable_init(cls):
        return cls(0, vm.BigInteger.zero())


class GasBonusState(serialization.ISerializable, Sequence):
    def __init__(self, initial_record: _GasRecord = None):
        self._storage_key = NeoToken().key_gas_per_block
        self._records: List[_GasRecord] = [initial_record] if initial_record else []
        self._iter = iter(self._records)

    def __len__(self):
        return len(self._records)

    def __iter__(self):
        self._iter = iter(self._records)
        return self

    def __next__(self) -> _GasRecord:
        return next(self._iter)

    def __getitem__(self, item):
        return self._records.__getitem__(item)

    def __setitem__(self, key, record: _GasRecord) -> None:
        self._records[key] = record

    @classmethod
    def from_snapshot(cls, snapshot: storage.Snapshot, read_only=False):
        storage_item = snapshot.storages.get(NeoToken().key_gas_per_block, read_only)
        return storage_item.get(cls)

    def append(self, record: _GasRecord) -> None:
        self._records.append(record)

    def serialize(self, writer: serialization.BinaryWriter) -> None:
        writer.write_serializable_list(self._records)

    def deserialize(self, reader: serialization.BinaryReader) -> None:
        self._records = reader.read_serializable_list(_GasRecord)


class NeoToken(FungibleToken):
    _id: int = -5
    _decimals: int = 0

    key_committee = storage.StorageKey(_id, b'\x0e')
    key_candidate = storage.StorageKey(_id, b'\x21')
    key_voters_count = storage.StorageKey(_id, b'\x01')
    key_gas_per_block = storage.StorageKey(_id, b'\x29')
    key_voter_reward_per_committee = storage.StorageKey(_id, b'\x17')
    key_register_price = storage.StorageKey(_id, b'\x0D')

    _NEO_HOLDER_REWARD_RATIO = 10
    _COMMITTEE_REWARD_RATIO = 10
    _VOTER_REWARD_RATIO = 80
    _symbol = "NEO"
    _state = _NeoTokenStorageState
    _candidates_dirty = True
    _candidates: List[Tuple[cryptography.ECPoint, vm.BigInteger]] = []

    def init(self):
        super(NeoToken, self).init()
        # singleton init, similar to __init__ but called only once
        self.total_amount = self.factor * 100_000_000
        self._committee_state = None

    def _initialize(self, engine: contracts.ApplicationEngine) -> None:
        # NEO's native contract initialize. Is called upon contract deploy

        self._committee_state = _CommitteeState(engine.snapshot,
                                                dict.fromkeys(settings.standby_validators, vm.BigInteger(0)))
        engine.snapshot.storages.put(self.key_voters_count, storage.StorageItem(b'\x00'))

        gas_bonus_state = GasBonusState(_GasRecord(0, GasToken().factor * 5))
        engine.snapshot.storages.put(self.key_gas_per_block, storage.StorageItem(gas_bonus_state.to_array()))
        engine.snapshot.storages.put(self.key_register_price,
                                     storage.StorageItem((GasToken().factor * 1000).to_array()))
        self.mint(engine,
                  contracts.Contract.get_consensus_address(settings.standby_validators),
                  self.total_amount,
                  False)

    @register("unclaimedGas", contracts.CallFlags.READ_STATES, cpu_price=1 << 17)
    def unclaimed_gas(self, snapshot: storage.Snapshot, account: types.UInt160, end: int) -> vm.BigInteger:
        """
        Return the available bonus GAS for an account.

        Requires sending the accounts balance to release the bonus GAS.

        Args:
            snapshot: snapshot of storage
            account: account to calculate bonus for
            end: ending block height to calculate bonus up to. You should use mostlikly use the current chain height.
        """
        storage_item = snapshot.storages.try_get(self.key_account + account, read_only=True)
        if storage_item is None:
            return vm.BigInteger.zero()
        state = storage_item.get(self._state)
        return self._calculate_bonus(snapshot, state.vote_to, state.balance, state.balance_height, end)

    @register("registerCandidate", contracts.CallFlags.STATES)
    def register_candidate(self,
                           engine: contracts.ApplicationEngine,
                           public_key: cryptography.ECPoint) -> bool:
        """
        Register a candidate for consensus node election.

        Args:
            engine: Application engine instance
            public_key: the candidate's public key

        Returns:
            True is succesfully registered. False otherwise.
        """
        script_hash = to_script_hash(contracts.Contract.create_signature_redeemscript(public_key))
        if not engine.checkwitness(script_hash):
            return False

        engine.add_gas(self.get_register_price(engine.snapshot))
        storage_key = self.key_candidate + public_key
        storage_item = engine.snapshot.storages.try_get(storage_key, read_only=False)
        if storage_item is None:
            state = _CandidateState()
            state.registered = True
            storage_item = storage.StorageItem(state.to_array())
            engine.snapshot.storages.put(storage_key, storage_item)
        else:
            state = storage_item.get(_CandidateState)
            state.registered = True
        self._candidates_dirty = True
        return True

    @register("unregisterCandidate", contracts.CallFlags.STATES, cpu_price=1 << 16)
    def unregister_candidate(self,
                             engine: contracts.ApplicationEngine,
                             public_key: cryptography.ECPoint) -> bool:
        """
        Remove a candidate from the consensus node candidate list.
        Args:
            engine: Application engine instance
            public_key: the candidate's public key

        Returns:
            True is succesfully removed. False otherwise.

        """
        script_hash = to_script_hash(contracts.Contract.create_signature_redeemscript(public_key))
        if not engine.checkwitness(script_hash):
            return False

        storage_key = self.key_candidate + public_key
        storage_item = engine.snapshot.storages.try_get(storage_key, read_only=False)
        if storage_item is None:
            return True
        else:
            state = storage_item.get(_CandidateState)
            state.registered = False
            if state.votes == 0:
                engine.snapshot.storages.delete(storage_key)

        self._candidates_dirty = True
        return True

    @register("vote", contracts.CallFlags.STATES, cpu_price=1 << 16)
    def vote(self,
             engine: contracts.ApplicationEngine,
             account: types.UInt160,
             vote_to: Optional[cryptography.ECPoint]) -> bool:
        """
        Vote on a consensus candidate
        Args:
            engine: Application engine instance.
            account: source account to take voting balance from
            vote_to: candidate public key

        Returns:
            True is vote registered succesfully. False otherwise.
        """
        if not engine.checkwitness(account):
            return False

        storage_key_account = self.key_account + account
        storage_item = engine.snapshot.storages.try_get(storage_key_account, read_only=False)
        if storage_item is None:
            return False
        account_state = storage_item.get(self._state)

        candidate_state = None
        if vote_to is not None:
            storage_key_candidate = self.key_candidate + vote_to
            storage_item_candidate = engine.snapshot.storages.try_get(storage_key_candidate, read_only=False)
            if storage_item_candidate is None:
                return False

            candidate_state = storage_item_candidate.get(_CandidateState)
            if not candidate_state.registered:
                return False

        if account_state.vote_to.is_zero() ^ (vote_to is None):
            si_voters_count = engine.snapshot.storages.get(self.key_voters_count, read_only=False)
            old_value = vm.BigInteger(si_voters_count.value)
            if account_state.vote_to.is_zero():
                new_value = old_value + account_state.balance
            else:
                new_value = old_value - account_state.balance
            si_voters_count.value = new_value.to_array()

        self._distribute_gas(engine, account, account_state)

        if not account_state.vote_to.is_zero():
            sk_validator = self.key_candidate + account_state.vote_to
            si_validator = engine.snapshot.storages.get(sk_validator, read_only=False)
            validator_state = si_validator.get(_CandidateState)
            validator_state.votes -= account_state.balance

            if not validator_state.registered and validator_state.votes == 0:
                reward_key = self.key_voter_reward_per_committee + account_state.vote_to.to_array()
                for k, _ in engine.snapshot.storages.find(reward_key):
                    engine.snapshot.storages.delete(k)
                engine.snapshot.storages.delete(sk_validator)

        account_state.vote_to = vote_to if vote_to else cryptography.ECPoint.deserialize_from_bytes(b'\x00')
        if candidate_state is not None:
            candidate_state.votes += account_state.balance
        self._candidates_dirty = True

        return True

    @register("getCandidates", contracts.CallFlags.READ_STATES, cpu_price=1 << 22)
    def get_candidates(self, engine: contracts.ApplicationEngine) -> List[Tuple[cryptography.ECPoint, vm.BigInteger]]:
        """
        Fetch all registered candidates, convert them to a StackItem and push them onto the evaluation stack.
        """
        return self._get_candidates(engine.snapshot)

    @register("getNextBlockValidators", contracts.CallFlags.READ_STATES, cpu_price=1 << 16)
    def get_next_block_validators(self, snapshot: storage.Snapshot) -> List[cryptography.ECPoint]:
        """
        Get the public keys of the consensus nodes that will validate the next block.
        """
        keys = self.get_committee_from_cache(snapshot)[:settings.network.validators_count]
        keys.sort()
        return keys

    @register("setGasPerBlock", contracts.CallFlags.STATES, cpu_price=1 << 15)
    def _set_gas_per_block(self, engine: contracts.ApplicationEngine, gas_per_block: vm.BigInteger) -> None:
        if gas_per_block > 0 or gas_per_block > 10 * self._gas.factor:
            raise ValueError("new gas per block value exceeds limits")

        if not self._check_committee(engine):
            raise ValueError("Check committee failed")

        index = engine.snapshot.persisting_block.index + 1
        gas_bonus_state = GasBonusState.from_snapshot(engine.snapshot, read_only=False)
        if gas_bonus_state[-1].index == index:
            gas_bonus_state[-1] = _GasRecord(index, gas_per_block)
        else:
            gas_bonus_state.append(_GasRecord(index, gas_per_block))

    @register("getGasPerBlock", contracts.CallFlags.READ_STATES, cpu_price=1 << 15)
    def get_gas_per_block(self, snapshot: storage.Snapshot) -> vm.BigInteger:
        """
        Get the amount of gas generated in each block.

        Args:
            snapshot: the snapshot to read the data from

        Returns:
            The amount of gas generated.
        """
        index = snapshot.best_block_height + 1
        gas_bonus_state = GasBonusState.from_snapshot(snapshot, read_only=True)
        for record in reversed(gas_bonus_state):  # type: _GasRecord
            if record.index <= index:
                return record.gas_per_block
        else:
            raise ValueError

    @register("setRegisterPrice", contracts.CallFlags.STATES, cpu_price=1 << 15)
    def set_register_price(self, engine: contracts.ApplicationEngine, register_price: int) -> None:
        if register_price <= 0:
            raise ValueError("Register price cannot be negative or zero")
        if not self._check_committee(engine):
            raise ValueError("CheckCommittee failed for setRegisterPrice")
        item = engine.snapshot.storages.get(self.key_register_price, read_only=False)
        item.value = vm.BigInteger(register_price).to_array()

    @register("getRegisterPrice", contracts.CallFlags.READ_STATES, cpu_price=1 << 15)
    def get_register_price(self, snapshot: storage.Snapshot) -> int:
        """
        Get the price for registering as a candidate
        """
        return int(vm.BigInteger(snapshot.storages.get(self.key_register_price, read_only=True).value))

    def _distribute_gas(self,
                        engine: contracts.ApplicationEngine,
                        account: types.UInt160,
                        state: _NeoTokenStorageState) -> None:
        if engine.snapshot.persisting_block is None:
            return

        gas = self._calculate_bonus(engine.snapshot, state.vote_to, state.balance, state.balance_height,
                                    engine.snapshot.persisting_block.index)
        state.balance_height = engine.snapshot.persisting_block.index
        GasToken().mint(engine, account, gas, True)

    @register("getCommittee", contracts.CallFlags.READ_STATES, cpu_price=1 << 16)
    def get_committee(self, snapshot: storage.Snapshot) -> List[cryptography.ECPoint]:
        """
        Get the public keys of the current validators.
        """
        return sorted(self.get_committee_from_cache(snapshot))

    @register("getAccountState", contracts.CallFlags.READ_STATES, cpu_price=1 << 15)
    def get_account_state(self, snapshot: storage.Snapshot, account: types.UInt160) -> Optional[_NeoTokenStorageState]:
        storage_key_account = self.key_account + account
        storage_item = snapshot.storages.try_get(storage_key_account, read_only=False)
        if storage_item is None:
            return None
        return storage_item.get(self._state)

    def total_supply(self, snapshot: storage.Snapshot) -> vm.BigInteger:
        """ Get the total deployed tokens. """
        return self.total_amount

    def on_balance_changing(self, engine: contracts.ApplicationEngine,
                            account: types.UInt160,
                            state,
                            amount: vm.BigInteger) -> None:
        self._distribute_gas(engine, account, state)

        if amount == vm.BigInteger.zero():
            return

        if state.vote_to.is_zero():
            return

        si_voters_count = engine.snapshot.storages.get(self.key_voters_count, read_only=False)
        new_value = vm.BigInteger(si_voters_count.value) + amount
        si_voters_count.value = new_value.to_array()

        si_candidate = engine.snapshot.storages.get(self.key_candidate + state.vote_to, read_only=False)
        candidate_state = si_candidate.get(_CandidateState)
        candidate_state.votes += amount
        self._candidates_dirty = True
        self._check_candidate(engine.snapshot, state.vote_to, candidate_state)

    def on_persist(self, engine: contracts.ApplicationEngine) -> None:
        super(NeoToken, self).on_persist(engine)

        # set next committee
        if self._should_refresh_committee(engine.snapshot.persisting_block.index):
            validators = self._compute_committee_members(engine.snapshot)
            if self._committee_state is None:
                self._committee_state = _CommitteeState.from_snapshot(engine.snapshot)
            self._committee_state._validators = validators
            self._committee_state.persist(engine.snapshot)

    def post_persist(self, engine: contracts.ApplicationEngine):
        super(NeoToken, self).post_persist(engine)
        # distribute GAS for committee
        m = len(settings.standby_committee)
        n = settings.network.validators_count
        index = engine.snapshot.persisting_block.index % m
        gas_per_block = self.get_gas_per_block(engine.snapshot)
        committee = self.get_committee_from_cache(engine.snapshot)
        pubkey = committee[index]
        account = to_script_hash(contracts.Contract.create_signature_redeemscript(pubkey))
        GasToken().mint(engine, account, gas_per_block * self._COMMITTEE_REWARD_RATIO / 100, False)

        if self._should_refresh_committee(engine.snapshot.persisting_block.index):
            voter_reward_of_each_committee = gas_per_block * self._VOTER_REWARD_RATIO * 100000000 * m / (m + n) / 100
            for i, member in enumerate(committee):
                factor = 2 if i < n else 1
                member_votes = self._committee_state[member]
                if member_votes > 0:
                    voter_sum_reward_per_neo = voter_reward_of_each_committee * factor / member_votes
                    voter_reward_key = (self.key_voter_reward_per_committee
                                        + member
                                        + self._to_uint32(engine.snapshot.persisting_block.index + 1)
                                        )
                    border = (self.key_voter_reward_per_committee + member).to_array()
                    try:
                        pair = next(engine.snapshot.storages.find_range(voter_reward_key.to_array(), border, "reverse"))
                        result = vm.BigInteger(pair[1].value)
                    except StopIteration:
                        result = vm.BigInteger.zero()
                    voter_sum_reward_per_neo += result
                    engine.snapshot.storages.put(voter_reward_key,
                                                 storage.StorageItem(voter_sum_reward_per_neo.to_array()))

    def get_committee_from_cache(self, snapshot: storage.Snapshot) -> List[cryptography.ECPoint]:
        if self._committee_state is None:
            self._committee_state = _CommitteeState.from_snapshot(snapshot)
        return self._committee_state.validators

    def get_committee_address(self, snapshot: storage.Snapshot) -> types.UInt160:
        """ Get the script hash of the current committee """
        comittees = self.get_committee(snapshot)
        return to_script_hash(
            contracts.Contract.create_multisig_redeemscript(
                len(comittees) - (len(comittees) - 1) // 2,
                comittees)
        )

    def _calculate_bonus(self,
                         snapshot: storage.Snapshot,
                         vote: cryptography.ECPoint,
                         value: vm.BigInteger,
                         start: int,
                         end: int) -> vm.BigInteger:
        if value == vm.BigInteger.zero() or start >= end:
            return vm.BigInteger.zero()

        if value.sign < 0:
            raise ValueError("Can't calculate bonus over negative balance")

        neo_holder_reward = self._calculate_neo_holder_reward(snapshot, value, start, end)
        if vote.is_zero():
            return neo_holder_reward
        border = (self.key_voter_reward_per_committee + vote).to_array()
        start_bytes = self._to_uint32(start)
        key_start = (self.key_voter_reward_per_committee + vote + start_bytes).to_array()

        try:
            pair = next(snapshot.storages.find_range(key_start, border, "reverse"))
            start_reward_per_neo = vm.BigInteger(pair[1].value)  # first pair returned, StorageItem
        except StopIteration:
            start_reward_per_neo = vm.BigInteger.zero()

        end_bytes = self._to_uint32(end)
        key_end = (self.key_voter_reward_per_committee + vote + end_bytes).to_array()

        try:
            pair = next(snapshot.storages.find_range(key_end, border, "reverse"))
            end_reward_per_neo = vm.BigInteger(pair[1].value)  # first pair returned, StorageItem
        except StopIteration:
            end_reward_per_neo = vm.BigInteger.zero()

        return neo_holder_reward + value * (end_reward_per_neo - start_reward_per_neo) / 100000000

    def _calculate_neo_holder_reward(self,
                                     snapshot: storage.Snapshot,
                                     value: vm.BigInteger,
                                     start: int,
                                     end: int) -> vm.BigInteger:
        gas_bonus_state = GasBonusState.from_snapshot(snapshot, read_only=True)
        gas_sum = 0
        for pair in reversed(gas_bonus_state):  # type: _GasRecord
            cur_idx = pair.index
            if cur_idx >= end:
                continue
            if cur_idx > start:
                gas_sum += pair.gas_per_block * (end - cur_idx)
                end = cur_idx
            else:
                gas_sum += pair.gas_per_block * (end - start)
                break
        return value * gas_sum * self._NEO_HOLDER_REWARD_RATIO / 100 / self.total_amount

    def _check_candidate(self,
                         snapshot: storage.Snapshot,
                         public_key: cryptography.ECPoint,
                         candidate: _CandidateState) -> None:
        if not candidate.registered and candidate.votes == 0:
            for k, v in snapshot.storages.find((self.key_voter_reward_per_committee + public_key).to_array()):
                snapshot.storages.delete(k)
            snapshot.storages.delete(self.key_candidate + public_key)

    def _compute_committee_members(self, snapshot: storage.Snapshot) -> Dict[cryptography.ECPoint, vm.BigInteger]:
        storage_item = snapshot.storages.get(self.key_voters_count, read_only=True)
        voters_count = int(vm.BigInteger(storage_item.value))
        voter_turnout = voters_count / float(self.total_amount)

        candidates = self._get_candidates(snapshot)
        if voter_turnout < 0.2 or len(candidates) < len(settings.standby_committee):
            results = {}
            for key in settings.standby_committee:
                for pair in candidates:
                    if pair[0] == key:
                        results.update({key: pair[1]})
                        break
                else:
                    results.update({key: vm.BigInteger.zero()})
            return results
        # first sort by votes descending, then by ECPoint ascending
        # we negate the value of the votes (c[1]) such that they get sorted in descending order
        candidates.sort(key=lambda c: (-c[1], c[0]))
        trimmed_candidates = candidates[:len(settings.standby_committee)]
        results = {}
        for candidate in trimmed_candidates:
            results.update({candidate[0]: candidate[1]})
        return results

    def _get_candidates(self,
                        snapshot: storage.Snapshot) -> \
            List[Tuple[cryptography.ECPoint, vm.BigInteger]]:
        self._candidates = []
        for k, v in snapshot.storages.find(self.key_candidate.to_array()):
            candidate = _CandidateState.deserialize_from_bytes(v.value)
            if candidate.registered:
                # take of the CANDIDATE prefix
                point = cryptography.ECPoint.deserialize_from_bytes(k.key[1:])
                self._candidates.append((point, candidate.votes))
        self._candidates_dirty = False
        return self._candidates

    def _should_refresh_committee(self, height: int) -> bool:
        return height % len(settings.standby_committee) == 0

    def _to_uint32(self, value: int) -> bytes:
        return struct.pack(">I", value)


class GasToken(FungibleToken):
    _id: int = -6
    _decimals: int = 8

    _state = FungibleTokenStorageState
    _symbol = "GAS"

    def _initialize(self, engine: contracts.ApplicationEngine) -> None:
        account = contracts.Contract.get_consensus_address(settings.standby_validators)
        self.mint(engine, account, vm.BigInteger(52_000_000) * self.factor, False)

    def on_persist(self, engine: contracts.ApplicationEngine) -> None:
        super(GasToken, self).on_persist(engine)
        total_network_fee = 0
        for tx in engine.snapshot.persisting_block.transactions:
            total_network_fee += tx.network_fee
            self.burn(engine, tx.sender, vm.BigInteger(tx.system_fee + tx.network_fee))
        pub_keys = NeoToken().get_next_block_validators(engine.snapshot)
        primary = pub_keys[engine.snapshot.persisting_block.primary_index]
        script_hash = to_script_hash(contracts.Contract.create_signature_redeemscript(primary))
        self.mint(engine, script_hash, vm.BigInteger(total_network_fee), False)



from __future__ import annotations
import asyncio
import traceback
from datetime import datetime
from neo3.network import payloads, convenience
from neo3.core import msgrouter
from neo3 import blockchain, network_logger as logger, _Singleton
from typing import Dict, Optional
from contextlib import suppress


class SyncManager(_Singleton):
    #: Maximum number of Blocks to cache in memory. Block persisting empties the cache allowing new blocks to be
    #: requested.
    BLOCK_MAX_CACHE_SIZE = 500
    #: Maximum number of blocks to ask per request. Cannot exceed
    #: :attr:`~neo3.network.payloads.block.GetBlockByIndexPayload.MAX_BLOCKS_COUNT`.
    BLOCK_NETWORK_REQ_LIMIT = 500
    #: Maximum time in seconds that a node may take to respond to a data request before it is tagged and eventually
    #: replaced.
    BLOCK_REQUEST_TIMEOUT = 5

    # init() is used instead of __init__() due to the Singleton inheritance (read its class documentation)
    def init(self):
        self.nodemgr = convenience.NodeManager()  # singleton
        self.ledger = blockchain.Blockchain()  # singleton
        self.block_cache = []
        self.block_requests = dict()  # type: Dict[int, convenience.RequestInfo]
        self.shutting_down = False

        self._is_persisting_blocks = False
        self._tasks = []
        self._service_task = None

        msgrouter.on_block += self.on_block_received

    async def start(self, timeout=5) -> None:
        """
        Start the block syncing service. Requires a started node manager.

        Args:
            timeout: time in seconds to wait for finding a started node manager.

        Raises:
            Exception: if no started Nodemanager is found within `timeout` seconds.
        """
        async def wait_for_nodemanager():
            logger.debug("Waiting for nodemanager to start")
            while not self.nodemgr.is_running:
                await asyncio.sleep(0.1)

        try:
            await asyncio.wait_for(wait_for_nodemanager(), timeout)
        except asyncio.TimeoutError:
            error_msg = f"Nodemanager failed to start within specified timeout {timeout}"
            logger.debug(error_msg)
            await self.shutdown()
            raise Exception(error_msg)

        logger.debug("Starting services")
        self._service_task = asyncio.create_task(self._run_service())

    async def shutdown(self) -> None:
        """
        Gracefully shutdown the sync manager.

        Stops block persisting and all service tasks.

        Note:
            This dependents on asyncio's Task canceling logic. It waits for all tasks to be cancelled and/or stopped
            before returning.
        """
        logger.debug("Syncmanager shutting down")
        self.shutting_down = True
        self.block_cache = []

        logger.debug("Stopping tasks...")
        if self._service_task:
            self._tasks.append(self._service_task)

        for t in self._tasks:
            t.cancel()
        await asyncio.gather(*self._tasks, return_exceptions=True)

    def on_block_received(self, from_nodeid, block: payloads.Block) -> int:
        try:
            request_info = self.block_requests.pop(block.index)
        except KeyError:
            # Received a block not asked for. Could be:
            # - a very late reply of a node, for which the block was already received and processed via another node
            # - rogue actors sending data
            return -1

        try:
            flight_info = request_info.flights.pop(from_nodeid)
        except KeyError:
            # received a target block from a node not asked from. We can't be sure this is valid
            # restore the request_info
            self.block_requests[block.index] = request_info
            return -2

        now = datetime.utcnow().timestamp()
        delta_time = now - flight_info.start_time
        speed = (len(block) / 1024) / delta_time  # KB/s

        node = self.nodemgr.get_node_by_nodeid(flight_info.node_id)
        if node:
            node.nodeweight.append_new_speed(speed)

        if not self._is_in_blockcache(block.index) and not self.shutting_down:
            self.block_cache.append(block)

        return 0

    async def _run_service(self) -> None:
        while not self.shutting_down:
            await self._check_timeout()
            await self._sync_blocks()
            if not self._is_persisting_blocks and len(self.block_cache) > 0:
                asyncio.create_task(self.persist_blocks())
            await asyncio.sleep(1)

    async def _check_timeout(self) -> int:
        """
        This function checks if any of the outstanding data requests have exceeded the response time threshold.
        If so then the violating node is tagged. Next, a new node is selected to request the data we have not yet
        received in the hope that this node does perform adequately.
        """
        if len(self.block_requests) == 0:
            # no outstanding data requests
            return -1

        timedout_flights = dict()
        now = datetime.utcnow().timestamp()

        # find outstanding requests that timed out
        for height, request_info in self.block_requests.items():
            flight_info = request_info.most_recent_flight()
            if flight_info and now - flight_info.start_time > self.BLOCK_REQUEST_TIMEOUT:
                timedout_flights[height] = flight_info

        if len(timedout_flights) == 0:
            # no timeouts, every request is still nicely within the threshold
            return -2

        remaining_requests = []
        nodes_to_tag_for_timeout = set()
        best_stored_block_height = self._get_best_stored_block_height()
        for height, flight_info in timedout_flights.items():
            # adding to set to ensure we only tag nodes once per request
            nodes_to_tag_for_timeout.add(flight_info.node_id)

            try:
                request_info = self.block_requests[height]
            except KeyError:
                # TODO: check if still possible. After refactor should not be reachable anymore
                continue

            if flight_info.height <= best_stored_block_height:
                with suppress(KeyError):
                    self.block_requests.pop(height)
                continue

            # tag the node for not delivering data within the set threshold
            request_info.mark_failed_node(flight_info.node_id)
            remaining_requests.append(request_info)

        for node_id in nodes_to_tag_for_timeout:
            # affect node weighting by increasing node timeout count
            self.nodemgr.increase_node_timeout_count(node_id)

        if len(remaining_requests) > 0:
            request_info_first = remaining_requests[0]
            request_info_last = remaining_requests[-1]
            # using the last request_info to find a suitable node, because the last request info is always the
            # highest block to look for
            node = self.nodemgr.get_least_failed_node(request_info_last)
            if node is None:
                # no connected nodes that can satisfy our request.
                # Return and let the node manager first resolve finding nodes
                return -3

            # it is only possible to request block data by height (using the GetBlockData payload) for a consecutive
            # range. One option is to find these ranges and send a request for each range. Another option, which keeps
            # the code much simpler, is to just request the full range (from start to end height) and ignore any gaps
            # in the range that have been filled in the mean time by other nodes that timed out.
            # This leads to minimal (acceptable) additional traffic in certain scenarios.
            for request_info in remaining_requests:
                request_info.add_new_flight(convenience.FlightInfo(node.nodeid, request_info.height))

            count = max(1, request_info_last.height - request_info_first.height)
            logger.debug(f"Block timeout for blocks {request_info_first.height} - {request_info_last.height}. "
                         f"Trying again using next available node {node.nodeid_human}. "
                         f"start={request_info_first.height}, count={count}.")
            await node.request_block_data(index_start=request_info_first.height, count=count)
            node.nodeweight.append_new_request_time()

        return 0

    async def persist_blocks(self) -> None:
        self._is_persisting_blocks = True
        self.block_cache.sort(key=lambda b: b.index)
        try:
            while not self.shutting_down:
                try:
                    block = self.block_cache.pop(0)
                except IndexError:
                    # cache empty
                    break
                msgrouter.on_block_persist(block)
                self.ledger.persist(block)
                await asyncio.sleep(0)
        except Exception as e:
            logger.debug(f"Unexpected exception happened while processing the block cache: {traceback.format_exc()}")
        finally:
            self._is_persisting_blocks = False

    async def _sync_blocks(self) -> int:
        # to simplify syncing, don't ask for more data when there are still requests in flight
        if len(self.block_requests) > 0:
            return -1

        block_cache_space = self.BLOCK_MAX_CACHE_SIZE - len(self.block_cache)
        if block_cache_space <= 0:
            return -2

        try:
            best_node_height = max(map(lambda node: node.best_height, self.nodemgr.nodes))
        except ValueError:
            # if the node list is empty max() fails on an empty list
            return -3

        node = self.nodemgr.get_node_with_height(best_node_height)
        # if not node:
        #     # no nodes with our desired height. We'll wait for node manager to resolve this
        #     # or for the nodes to increase their height on the next produced block
        #     return -4

        best_block_height = self._get_best_stored_block_height()
        block_request_limit = min(block_cache_space, self.BLOCK_NETWORK_REQ_LIMIT)

        to_fetch_ctr = 0
        for i in range(1, block_request_limit + 1):
            next_block_height = best_block_height + i

            if next_block_height > best_node_height:
                break

            self._add_block_flight_info(node.nodeid, next_block_height)
            to_fetch_ctr += 1

        if to_fetch_ctr > 0:
            index_start = best_block_height + 1
            logger.debug(f"Asking for blocks {index_start} - {index_start + to_fetch_ctr - 1} from {node.nodeid_human} "
                         f"(blocks in cache: {len(self.block_cache)}).")
            await node.request_block_data(index_start=index_start, count=to_fetch_ctr)

        return 0

    def _get_best_stored_block_height(self) -> int:
        """
        Helper to return the highest block in our possession (either in ledger or in block_cache)
        """
        best_block_cache_height = -1
        self.block_cache.sort(key=lambda b: b.index)
        if len(self.block_cache) > 0:
            best_block_cache_height = self.block_cache[-1].index

        ledger_height = self.ledger.height

        return max(ledger_height, best_block_cache_height)

    def _add_block_flight_info(self, nodeid, height: int) -> None:
        request_info = self.block_requests.get(height, None)

        if request_info is None:
            # no outstanding requests for this particular height, so we create it
            req = convenience.RequestInfo(height)
            req.add_new_flight(convenience.FlightInfo(nodeid, height))
            self.block_requests[height] = req
        else:
            request_info.flights.update({nodeid: convenience.FlightInfo(nodeid, height)})

    def _is_in_blockcache(self, block_height: int) -> bool:
        for b in self.block_cache:
            if b.index == block_height:
                return True
        else:
            return False

    def _run_in_loop(self, coro, interval) -> None:
        """
        Helper function to run a coroutine every `interval` seconds as long as the sync manager is running.

        Stores a task in `_tasks`

        Args:
            coro:
            interval:
        """

        async def _t(coro, interval):
            while not self.shutting_down:
                await coro()
                await asyncio.sleep(interval)

        self._tasks.append(asyncio.create_task(_t(coro, interval)))

    def _reset_for_test(self) -> None:
        self.block_cache = []
        self.block_requests = dict()
        self.shutting_down = False

        self._is_persisting_blocks = False
        self._tasks = []
        self._service_task = None



from __future__ import annotations
from enum import IntEnum, auto
from typing import Iterator, List
import json


class JsonPathType(IntEnum):
    ROOT = auto()
    DOT = auto()
    LEFT_BRACKET = auto()
    RIGHT_BRACKET = auto()
    ASTERISK = auto()
    COMMA = auto()
    COLON = auto()
    IDENTIFIER = auto()
    STRING = auto()
    NUMBER = auto()


class FormatException(ValueError):
    pass


class MaxDepthException(Exception):
    pass


class JsonPathToken:
    def __init__(self):
        self.type = None
        self.content: str = ''

    def __len__(self):
        try:
            return len(self.content)
        except TypeError:
            return 0

    def __repr__(self):
        return f"<JsonPathToken @ {hex(id(self))}> {self.type.name}: {self.content}"

    @staticmethod
    def parse(expression: str) -> Iterator[JsonPathToken]:
        i = 0
        while i < len(expression):
            token = JsonPathToken()
            c = expression[i]
            if c == '$':
                token.type = JsonPathType.ROOT
            elif c == '.':
                token.type = JsonPathType.DOT
            elif c == '[':
                token.type = JsonPathType.LEFT_BRACKET
            elif c == ']':
                token.type = JsonPathType.RIGHT_BRACKET
            elif c == '*':
                token.type = JsonPathType.ASTERISK
            elif c == ',':
                token.type = JsonPathType.COMMA
            elif c == ':':
                token.type = JsonPathType.COLON
            elif c == '\'':
                token.type = JsonPathType.STRING
                token.content = JsonPathToken.parse_string(expression, i)
                i += len(token) - 1
            elif c == '_' or c.isalpha():
                token.type = JsonPathType.IDENTIFIER
                token.content = JsonPathToken.parse_identifier(expression, i)
                i += len(token) - 1
            elif c == '-' or c.isdigit():
                token.type = JsonPathType.NUMBER
                token.content = JsonPathToken.parse_number(expression, i)
                i += len(token) - 1
            else:
                raise FormatException("Invalid token")

            yield token
            i += 1

    @staticmethod
    def parse_string(expression: str, start: int) -> str:
        for i, c in enumerate(expression[start:]):
            if c == '\'':
                return expression[start:start + i]
        else:
            raise FormatException("No closing \' found")

    @staticmethod
    def parse_identifier(expression: str, start: int) -> str:
        for i, c in enumerate(expression[start:]):
            if c != '_' and not c.isalpha() and not c.isdigit():
                return expression[start:start + i]
        else:
            return expression[start:]

    @staticmethod
    def parse_number(expression: str, start: int) -> str:
        for i, c in enumerate(expression[start + 1:]):
            if not c.isdigit():
                return expression[start:start + i + 1]
        else:
            return expression[start:]


class JsonPath:
    def __init__(self, json_: dict):
        self.json = json_
        self.max_depth = -1
        self.depth = -1

    def next_token(self):
        return next(self.tokens)

    def parse(self, expression: str, max_depth=6):
        if len(expression) == 0:
            return self.json
        self.max_depth = self.depth = max_depth
        self.tokens = JsonPathToken.parse(expression)
        first = self.next_token()
        if first.type != JsonPathType.ROOT:
            raise FormatException("First token must be ROOT ($)")

        objects: List[object] = [self.json]
        for token in self.tokens:
            if token.type == JsonPathType.DOT:
                objects = self.process_dot(objects)
            elif token.type == JsonPathType.LEFT_BRACKET:
                objects = self.process_left_bracket(objects)
        return objects

    def process_dot(self, objects: List[object]):
        token = self.next_token()
        if token.type == JsonPathType.ASTERISK:
            return self.descent(objects)
        elif token.type == JsonPathType.DOT:
            return self.descent_recursive(objects)
        elif token.type == JsonPathType.IDENTIFIER:
            return self.descent_identifier(objects, [token.content])
        else:
            raise FormatException("Invalid token")

    def descent(self, objects):
        self._check_depth()
        values = []
        for obj in objects:
            if isinstance(obj, list):
                values.extend(obj)
            if isinstance(obj, dict):
                keys = sorted(obj.keys(), reverse=True)
                for k in keys:
                    values.append(obj[k])
        return values

    def descent_recursive(self, objects):
        token = self.next_token()
        if token.type != JsonPathType.IDENTIFIER:
            raise FormatException(f"Expected IDENTIFIER got {token.type.name}")

        values = []
        while len(objects) > 0:
            values.extend(self.descent_identifier(objects, [token.content], check_depth=False))
            objects = self.descent(objects)
        return values

    def descent_identifier(self, objects, names: List[str], check_depth=True):
        if check_depth:
            self._check_depth()
        values = []
        for obj in objects:
            if not isinstance(obj, dict):
                continue
            for name in names:
                if name in obj:
                    values.append(obj[name])
        return values

    def process_left_bracket(self, objects):
        token = self.next_token()
        if token.type == JsonPathType.ASTERISK:
            token = self.next_token()
            if token.type != JsonPathType.RIGHT_BRACKET:
                raise FormatException(f"Only RIGHT_BRACKET is valid after `[*` found {token.type.name}")
            return self.descent(objects)
        elif token.type == JsonPathType.COLON:
            return self.process_slice(objects, 0)
        elif token.type == JsonPathType.NUMBER:
            next = self.next_token()
            if next.type == JsonPathType.COLON:
                return self.process_slice(objects, int(token.content))
            elif next.type == JsonPathType.COMMA:
                return self.process_union(objects, token)
            elif next.type == JsonPathType.RIGHT_BRACKET:
                return self.descent_index(objects, [int(token.content)])
            else:
                raise FormatException("Bracket group parsing failure for bracket with number")
        elif token.type == JsonPathType.STRING:
            next = self.next_token()
            if next.type == JsonPathType.COMMA:
                return self.process_union(objects, token)
            elif next.type == JsonPathType.RIGHT_BRACKET:
                return self.descent_identifier(objects, token.content)
            else:
                raise FormatException("Bracket group parsing failure for bracket with string")
        return objects

    def process_slice(self, objects, start: int):
        token = self.next_token()
        if token.type == JsonPathType.NUMBER:
            next = self.next_token()
            if next.type != JsonPathType.RIGHT_BRACKET:
                raise FormatException()
            return self.descent_range(objects, start, int(token.content))
        elif token.type == JsonPathType.RIGHT_BRACKET:
            return self.descent_range(objects, start, 0)
        else:
            raise FormatException()

    def process_union(self, objects, first_token):
        items = [first_token]
        while True:
            token = self.next_token()
            if token.type != first_token.type:
                raise FormatException()
            items.append(token)

            token = self.next_token()
            if token.type == JsonPathType.RIGHT_BRACKET:
                break
            if token.type != JsonPathType.COMMA:
                raise FormatException()

        if first_token.type == JsonPathType.NUMBER:
            items = list(map(lambda i: int(i.content), items))
            return self.descent_index(objects, items)
        elif first_token.type == JsonPathType.STRING:
            for item in items:
                item.content = item.content.replace("'", "")
            return self.descent_identifier(objects, items)
        else:
            raise FormatException()

    def descent_index(self, objects, indices: List[int]):
        self._check_depth()
        values = []
        for obj in objects:
            if not isinstance(obj, list):
                continue
            for idx in indices:
                if idx < 0:
                    idx += len(obj)
                if 0 <= idx < len(obj):
                    values.append(obj[idx])
        return values

    def descent_range(self, objects, start: int, end: int):
        self._check_depth()
        values = []
        for obj in objects:
            if not isinstance(obj, list):
                continue
            sub_start = start
            if sub_start < 0:
                sub_start += len(obj)

            sub_end = end
            if sub_end <= 0:
                sub_end += len(obj)

            if sub_end > len(obj):
                sub_end = len(obj)

            if sub_end - sub_start < 0:
                raise FormatException()

            values.extend(obj[sub_start:sub_end])

        return values

    def _check_depth(self):
        if self.depth <= 0:
            raise MaxDepthException()
        self.max_depth -= 1



"""
Courtesy of Guido: https://www.python.org/download/releases/2.2/descrintro/#__new__

To create a singleton class, you subclass from Singleton; each subclass will have a single instance,
no matter how many times its constructor is called. To further initialize the subclass instance, subclasses should
override 'init' instead of __init__ - the __init__ method is called each time the constructor is called.
"""


class _Singleton(object):
    def __new__(cls, *args, **kwds):
        it = cls.__dict__.get("__it__")
        if it is not None:
            return it
        cls.__it__ = it = object.__new__(cls)
        it.init(*args, **kwds)
        return it

    def init(self, *args, **kwds):
        """ override this instead of implementing your own __init__ """



from __future__ import annotations
from typing import Any, Dict, List, Optional
from jsonschema import validate  # type: ignore
from neo3.core import IJson, cryptography
from neo3.wallet.account import Account, AccountContract
from neo3 import contracts
from neo3.wallet.scrypt_parameters import ScryptParameters


class MultiSigContext:
    def __init__(self):
        self.initialised = False
        self.signing_threshold = 999
        #: List of valid public keys for signing
        self.expected_public_keys: List[cryptography.ECPoint] = []
        #: Completed pairs
        self.signature_pairs: Dict[cryptography.ECPoint, bytes] = {}

    @property
    def is_complete(self):
        return len(self.signature_pairs) >= self.signing_threshold

    def signing_status(self) -> Dict[cryptography.ECPoint, bool]:
        # shows which keys have been completed
        pass

    def process_contract(self, script: bytes) -> None:
        valid, threshold, public_keys = contracts.Contract.parse_as_multisig_contract(script)
        if not valid:
            raise ValueError("Invalid script")
        self.expected_public_keys = public_keys
        self.signing_threshold = threshold
        self.initialised = True


class Wallet(IJson):
    _wallet_version = '1.0'

    # Wallet JSON validation schema
    json_schema = {
        "type": "object",
        "properties": {
            "name": {"type": ["string", "null"]},
            "version": {"type": "string"},
            "scrypt": ScryptParameters.json_schema,
            "accounts": {
                "type": "array",
                "items": Account._json_schema,
                "minItems": 0,
            },
            "extra": {"type": ["object", "null"],
                      "properties": {},
                      "additionalProperties": True
                      },
        },
        "required": ["name", "version", "scrypt", "accounts", "extra"]
    }

    def __init__(self,
                 name: Optional[str] = None,
                 version: str = _wallet_version,
                 scrypt: Optional[ScryptParameters] = None,
                 accounts: List[Account] = None,
                 default_account: Optional[Account] = None,
                 extra: Optional[Dict[Any, Any]] = None):
        """
        Args:
            name: a user defined label for the wallet
            version: the wallet's version, must be equal to or greater than 3.0
            scrypt:  the parameters of the Scrypt algorithm used for encrypting and decrypting the private keys in the
            wallet.
            accounts: an array of Account objects to add to the wallet.
            extra: a user defined object for storing extra data. This field can be None.
        """

        self.name = name
        self.version = version
        self.scrypt = scrypt if scrypt else ScryptParameters()

        if accounts is None:
            accounts = []

        if default_account is not None and default_account not in accounts:
            # default account must be in the account list
            accounts.append(default_account)
        elif default_account is None and len(accounts) > 0:
            # if no account is defined as default, the first will be considered the default account
            default_account = accounts[0]

        self.accounts = accounts if accounts is not None else []
        self._default_account: Optional[Account] = default_account
        self.extra = extra if extra else {}

    def account_new(self, password: str, label: str = None, is_default=False) -> Account:
        """
        Creates a new account and adds it in the wallet

        Args:
            password: the password to encrypt the account
            label: optional label to identify the account
            is_default: whether it should set the created account as the default
        """
        account = Account(password=password,
                          watch_only=False,
                          label=label
                          )

        self.account_add(account, is_default)
        return account

    @property
    def account_default(self) -> Optional[Account]:
        """
        Returns the default account if at least one account is present
        """
        return self._default_account

    def import_multisig_address(self,
                                signing_threshold: int,
                                public_keys: List[cryptography.ECPoint]
                                ) -> Account:
        if signing_threshold < 1 or signing_threshold > 1024:
            raise ValueError("Invalid signing threshold")

        if signing_threshold > len(public_keys):
            raise ValueError(f"Minimum signing threshold is {signing_threshold}, "
                             f"received only {len(public_keys)} public keys")

        multisig_contract = contracts.Contract.create_multisig_contract(signing_threshold, public_keys)
        # we start with a watchonly as base
        account = Account.watch_only(multisig_contract.script_hash)
        account.contract = AccountContract.from_contract(multisig_contract)

        # if the wallet contains an account matching one required in the multisig, then copy key material
        self._augment_multisig_with_key_material(account)
        self.account_add(account)
        return account

    def _augment_multisig_with_key_material(self, account: Account):
        """
        Tries to augment multisig accounts with key material such that they can be used for signing

        There are 2 scenario's
        1. A multisig account is added while there already exists a regular account with key material for one of the
        keys required by the multisig
        2. A regular account is added while there already exists a multisig account missing key material that the
        regular now adds.
        """
        if account.contract is None:
            return

        is_multisig, _, public_keys = contracts.Contract.parse_as_multisig_contract(account.contract.script)
        if is_multisig:
            # scenario 1
            for acc in self.accounts:
                if not acc.is_watchonly and acc.public_key in public_keys:
                    # copy key information of the first matching account
                    account.encrypted_key = acc.encrypted_key
                    account.public_key = acc.public_key
                    break
        else:
            # scenario 2
            for acc in self.accounts:
                if acc.is_watchonly and acc.is_multisig and acc.contract:  # testing acc.contract to silence mypy
                    _, _, public_keys = contracts.Contract.parse_as_multisig_contract(acc.contract.script)
                    if account.public_key in public_keys:
                        acc.encrypted_key = account.encrypted_key
                        acc.public_key = account.public_key
                        break

    def account_add(self, account: Account, is_default=False) -> bool:
        """
        Includes an account in the wallet

        Args:
            account: the account to be included
            is_default: whether it should set the created account as the default

        Raises:
            ValueError: if the account's label is already used by another one
        """
        # true if ok, false if duplicate (any other possible reasons? otherwise we need to throw exceptions)
        if account in self.accounts:
            return False

        if account.label is not None and self.account_get_by_label(account.label) is not None:
            raise ValueError(f"Label is already used by an account '{account.label}'")

        # if first account, also set to default
        if is_default or len(self.accounts) == 0:
            self._default_account = account

        self._augment_multisig_with_key_material(account)
        self.accounts.append(account)
        return True

    def account_delete(self, account: Account) -> bool:
        """
        Removes an account from the wallet

        Args:
            account: the account to be removed
        """
        # return success or not
        if account not in self.accounts:
            return False

        self.accounts.remove(account)
        if account == self._default_account:
            # if it was the default account, select a new one
            # first by default
            self._default_account = self.accounts[0] if len(self.accounts) > 0 else None

        return True

    def account_delete_by_label(self, label: str) -> bool:
        """
        Removes an account from the wallet given its label

        Args:
            label: unique identifier of the account
        """
        # return success or not
        account = self.account_get_by_label(label)

        if account is None:
            # account with that label was not found
            return False

        return self.account_delete(account)

    def account_get_by_label(self, label: str) -> Optional[Account]:
        """
        Gets an account given its label. Returns None if not found.

        Args:
            label: unique identifier of the account
        """
        # returns the account with given label. None if the account is not found
        return next((acc for acc in self.accounts if acc.label == label), None)

    def save(self):
        """
        Saves the wallet.

        This is called automatically when using the context manager.

        See Also:
            :class:`~neo3.wallet.nep6.nep6diskwallet.NEP6DiskWallet`
        """
        pass

    def to_json(self) -> dict:
        """
        Convert object into JSON representation.
        """
        return {
            'name': self.name,
            'version': self.version,
            'scrypt': self.scrypt.to_json(),
            'accounts': [self._account_to_json(account) for account in self.accounts],
            'extra': self.extra if len(self.extra) > 0 else None
        }

    def _account_to_json(self, account: Account) -> dict:
        is_default = self._default_account is not None and self._default_account.address == account.address
        json_account = account.to_json()
        json_account['isDefault'] = is_default
        return json_account

    @classmethod
    def from_json(cls, json: dict, password: str = None) -> Wallet:
        """
        Parse object out of JSON data.

        Args:
            json: a dictionary.
            password: the password to decrypt the json data.

        Raises:
            KeyError: if the data supplied does not contain the necessary keys.
            ValueError: if the 'version' property is under 1.0 or is not a valid string.
        """
        validate(json, schema=cls.json_schema)
        try:
            if float(json['version']) < 1.0:
                raise ValueError("Format error - invalid 'version'")
        except ValueError:
            raise ValueError("Format error - invalid 'version'")

        accounts = []
        default_account = None
        if len(json['accounts']) > 0:
            if password is None:
                raise ValueError('Missing wallet password to decrypt account data')
            else:
                for json_account in json['accounts']:
                    account_from_json = Account.from_json(json_account, password)
                    accounts.append(account_from_json)
                    if default_account is None and hasattr(json, 'isDefault') and json['isDefault']:
                        default_account = account_from_json

        return cls(name=json['name'],
                   version=json['version'],
                   scrypt=ScryptParameters.from_json(json['scrypt']),
                   accounts=accounts,
                   default_account=default_account,
                   extra=json['extra'])

    def __enter__(self) -> Wallet:
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        self.save()



from .decorator import register
from .nativecontract import NativeContract
from .fungible import (FungibleToken, NeoToken, GasToken, FungibleTokenStorageState)
from .policy import PolicyContract
from .oracle import OracleContract
from .designate import DesignateRole, DesignationContract
from .management import ManagementContract
from .ledger import LedgerContract
from .crypto import CryptoContract
from .stdlib import StdLibContract

__all__ = ['NativeContract',
           'PolicyContract',
           'NeoToken',
           'GasToken',
           'OracleContract',
           'DesignationContract',
           'ManagementContract',
           'LedgerContract',
           'FungibleToken',
           'CryptoContract',
           'StdLibContract',
           'register'
           ]



import unittest
from neo3.network import convenience


class RequestInfoTestCase(unittest.TestCase):
    def test_most_recent_flight(self):
        ri = convenience.RequestInfo(0)
        self.assertIsNone(ri.most_recent_flight())

        fi = convenience.FlightInfo(1, 0)
        ri.add_new_flight(fi)
        most_recent = ri.most_recent_flight()
        self.assertEqual(ri.last_used_node, fi.node_id)

    def test_mark_failed(self):
        ri = convenience.RequestInfo(0)
        self.assertEqual(0, ri.failed_total)
        self.assertEqual(0, len(ri.failed_nodes))

        ri.mark_failed_node(123)
        self.assertEqual(1, ri.failed_total)
        self.assertEqual(1, len(ri.failed_nodes))


class NodeWeightTestCase(unittest.TestCase):
    def test_weight(self):
        nw1 = convenience.NodeWeight(node_id=123)
        self.assertEqual(convenience.NodeWeight.SPEED_RECORD_COUNT, len(nw1.speed))
        nw1.append_new_speed(1)
        self.assertEqual(convenience.NodeWeight.SPEED_RECORD_COUNT, len(nw1.speed))

        nw2 = convenience.NodeWeight(node_id=456)
        nw2.append_new_speed(1000)

        # highest speed + longest time since used has best weight. Here nw1 has the worst speed,
        # but longest time since use. Therefore NW2 should win
        self.assertTrue(nw2 > nw1)

        # now make nw1 fastest, but test for being punished hard for timeouts
        nw1.append_new_speed(100_000)
        nw1.append_new_speed(100_000)
        nw1.append_new_speed(100_000)
        self.assertTrue(nw1 > nw2)
        nw1.timeout_count += 1
        self.assertFalse(nw1 > nw2)

        self.assertIn("<NodeWeight at", str(nw1))
        # for some reason the weight format sometimes rounds differently causing the test to fail, so we limit decimals to 0
        self.assertIn(f"weight:{nw1.weight():.0f}", str(nw1))



from .account import Account
from .scrypt_parameters import ScryptParameters
from .utils import calculate_system_fee, calculate_network_fee, add_network_fee, add_system_fee
from .wallet import Wallet, MultiSigContext



from .base import (IDBImplementation,
                   StorageContext)
from .storageitem import StorageItem
from .storagekey import StorageKey
from .cache import (Trackable,
                    TrackState,
                    CachedTXAccess,
                    CloneTXCache,
                    CachedBlockAccess,
                    CloneBlockCache,
                    CachedContractAccess,
                    CloneContractCache,
                    CachedStorageAccess,
                    CloneStorageCache,
                    AttributeCache)
from .snapshot import CloneSnapshot, Snapshot
from .utils import NEOByteCompare, NEOSeekSort, create_find_prefix



from __future__ import annotations
from typing import List
from neo3 import vm, contracts
from neo3.core import cryptography, types, to_script_hash
from neo3.contracts.interop import register


@register("System.Contract.Call", 1 << 15, contracts.CallFlags.READ_STATES | contracts.CallFlags.ALLOW_CALL)
def contract_call(engine: contracts.ApplicationEngine,
                  contract_hash: types.UInt160,
                  method: str,
                  call_flags: contracts.CallFlags,
                  args: vm.ArrayStackItem) -> None:
    if method.startswith("_"):
        raise ValueError("Invalid method name")

    target_contract = contracts.ManagementContract().get_contract(engine.snapshot, contract_hash)
    if target_contract is None:
        raise ValueError("[System.Contract.Call] Can't find target contract")

    method_descriptor = target_contract.manifest.abi.get_method(method, len(args))
    if method_descriptor is None:
        raise ValueError(f"[System.Contract.Call] Method '{method}' does not exist on target contract")

    has_return_value = method_descriptor.return_type != contracts.ContractParameterType.VOID
    if not has_return_value:
        engine.current_context.evaluation_stack.push(vm.NullStackItem())

    engine._contract_call_internal2(target_contract, method_descriptor, call_flags, has_return_value, list(args))


@register("System.Contract.GetCallFlags", 1 << 10, contracts.CallFlags.NONE)
def get_callflags(engine: contracts.ApplicationEngine) -> contracts.CallFlags:
    return contracts.CallFlags(engine.current_context.call_flags)


@register("System.Contract.CreateStandardAccount", 1 << 8, contracts.CallFlags.NONE)
def contract_create_standard_account(engine: contracts.ApplicationEngine,
                                     public_key: cryptography.ECPoint) -> types.UInt160:
    return to_script_hash(contracts.Contract.create_signature_redeemscript(public_key))


@register("System.Contract.CreateMultisigAccount", 1 << 8, contracts.CallFlags.NONE)
def contract_create_multisigaccount(engine: contracts.ApplicationEngine,
                                    m: int,
                                    public_keys: List[cryptography.ECPoint]) -> types.UInt160:
    return to_script_hash(contracts.Contract.create_multisig_redeemscript(m, public_keys))


@register("System.Contract.NativeOnPersist", 0, contracts.CallFlags.STATES)
def native_on_persist(engine: contracts.ApplicationEngine) -> None:
    if engine.trigger != contracts.TriggerType.ON_PERSIST:
        raise SystemError()
    # NEO implicitely expects the ManagementContract to be called first *ugh*
    # because ManagementContract.on_persist will call _initialize() on all other native contracts
    # which is needed for the other contracts to work properly when their on_persist() is called
    sorted_contracts = sorted(contracts.NativeContract().registered_contracts, key=lambda c: c.id, reverse=True)
    for contract in sorted_contracts:
        if contract.active_block_index <= engine.snapshot.persisting_block.index:
            contract.on_persist(engine)


@register("System.Contract.NativePostPersist", 0, contracts.CallFlags.STATES)
def native_post_persist(engine: contracts.ApplicationEngine) -> None:
    if engine.trigger != contracts.TriggerType.POST_PERSIST:
        raise SystemError()
    for contract in contracts.NativeContract._contracts.values():
        if contract.active_block_index <= engine.snapshot.persisting_block.index:
            contract.post_persist(engine)


@register("System.Contract.CallNative", 0, contracts.CallFlags.NONE)
def call_native(engine: contracts.ApplicationEngine, version: int) -> None:
    contract = contracts.NativeContract.get_contract_by_hash(engine.current_scripthash)
    if contract is None:
        raise ValueError(f"It is not allowed to use \"System.Contract.CallNative\" directly")
    if contract.active_block_index > engine.snapshot.persisting_block.index:
        raise ValueError(f"The native contract {contract.service_name()} is not active")
    contract.invoke(engine, version)



import mmh3  # type: ignore
from neo3.core import serialization, types, Size as s, utils
from neo3 import vm


class StorageKey(serialization.ISerializable):
    def __init__(self, id_: int, key: bytes):
        self.id = id_
        self.key = key

    def __len__(self):
        return s.uint32 + len(self.key)

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        return self.id == other.id and self.key == other.key

    def __hash__(self):
        return hash(self.id) + mmh3.hash(self.key, seed=0, signed=False)

    def __repr__(self):
        return f"<{self.__class__.__name__} at {hex(id(self))}> [{self.id}] {self.key}"

    def __add__(self, other):
        if type(other) in [bytes, bytearray]:
            return StorageKey(self.id, self.key + other)
        if isinstance(other, (serialization.ISerializable, vm.BigInteger)):
            return StorageKey(self.id, self.key + other.to_array())
        else:
            return NotImplemented

    def serialize(self, writer: serialization.BinaryWriter) -> None:
        """
        Serialize the object into a binary stream.

        Args:
            writer: instance.
        """
        writer.write_int32(self.id)
        writer.write_bytes(self.key)

    def deserialize(self, reader: serialization.BinaryReader) -> None:
        """
        Deserialize the object from a binary stream.

        Args:
            reader: instance.
        """
        self.id = reader.read_int32()
        remaining_stream_size = len(reader) - reader._stream.tell()
        self.key = reader.read_bytes(remaining_stream_size)

    @classmethod
    def _serializable_init(cls):
        return cls(0, b'')



from __future__ import annotations
import sys
import abc
import struct
from io import BytesIO, SEEK_END
from typing import Union, Any, List, Type, TypeVar


ISerializable_T = TypeVar('ISerializable_T', bound='ISerializable')

__all__ = ['ISerializable', 'BinaryReader', 'BinaryWriter']


class ISerializable(abc.ABC):
    """
    An interface like class supporting NEO's network serialization protocol.
    """
    @abc.abstractmethod
    def serialize(self, writer: BinaryWriter) -> None:
        """
        Serialize the object into a binary stream.

        Args:
            writer: instance.
        """

    @abc.abstractmethod
    def deserialize(self, reader: BinaryReader) -> None:
        """
        Deserialize the object from a binary stream.

        Args:
            reader: instance.
        """

    @classmethod
    def deserialize_from_bytes(cls: Type[ISerializable_T], data: Union[bytes, bytearray]) -> ISerializable_T:
        """
        Parse data into an object instance.

        Args:
            data: hex escaped bytes.

        Returns:
            a deserialized instance of the class.
        """
        with BinaryReader(data) as br:
            payload = cls._serializable_init()
            payload.deserialize(br)
            return payload

    def to_array(self) -> bytes:
        """ Serialize the object into a bytearray."""
        with BinaryWriter() as bw:
            self.serialize(bw)
            return bw._stream.getvalue()

    @abc.abstractmethod
    def __len__(self):
        """ Return the length of the object in number of bytes."""

    @classmethod
    def _serializable_init(cls):
        """
        If the interface inheritor has mandatory arguments, override this functin and provide dummy values. These values
        will be overwritten by the read_serializable, read_serializable_list and deserialize_from_bytes methods that
        rely on this function for class instantiation.
        """
        return cls()


class BinaryReader(object):
    _uint8 = struct.Struct("<B")
    _uint16 = struct.Struct("<H")
    _uint32 = struct.Struct("<I")
    _uint64 = struct.Struct("<Q")
    _int16 = struct.Struct("<h")
    _int32 = struct.Struct("<i")
    _int64 = struct.Struct("<q")
    _uint16BE = struct.Struct(">H")
    _uint32BE = struct.Struct(">I")
    _uint64BE = struct.Struct(">Q")
    _int16BE = struct.Struct(">h")
    _int32BE = struct.Struct(">i")
    _int64BE = struct.Struct(">q")
    """
        A convenience class for reading data from byte streams.

        Context manager support is available to ensure proper cleanup of resources.

        Example:
        ::

            with BinaryReader(b'\\x01\\x02') as br:
                my_value = br.read_uint16()
    """

    def __init__(self, stream: Union[bytes, bytearray]) -> None:
        """
        Create an instance.

        Args:
            stream: a stream to operate on.
        """
        super(BinaryReader, self).__init__()
        self._stream = BytesIO(stream)

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        self.close()

    def __len__(self):
        io = self._stream
        # Remember our current position
        cur_pos = io.tell()
        # Seek to the end of the File object
        io.seek(0, SEEK_END)
        # Remember position, which is equal to the full length
        full_size = io.tell()
        # Seek back to the current position
        io.seek(cur_pos)
        return full_size

    def read_byte(self) -> bytes:
        """
        Read a single byte.

        Raises:
            ValueError: if 1 byte of data cannot be read from the stream.

        Returns:
            bytes: a hex escaped bytearray with 1 element.
        """
        value = self._stream.read(1)
        if len(value) != 1:
            raise ValueError("Could not read byte from empty stream")
        return value

    def read_bytes(self, length: int, _skip_length_check: bool = False) -> bytes:
        """
        Read the specified number of bytes from the stream.

        Args:
            length: number of bytes to read.

        Raises:
            ValueError: if `length` bytes of data cannot be read from the stream.

        Returns:
            bytes: `length` number of bytes.
        """
        value = self._stream.read(length)
        if not _skip_length_check and len(value) != length:
            raise ValueError(f"Could not read {length} bytes from stream. Only found {len(value)} bytes of data")

        return value

    def read_bool(self) -> bool:
        """
        Read 1 byte as a boolean value from the stream.

        Returns:
            bool: False for b'\x00'. True for all other values.
        """
        return struct.unpack("?", self._stream.read(1))[0]

    def read_uint8(self) -> int:
        """
        Read 1 byte as an unsigned integer value from the stream.
        """
        return self._uint8.unpack(self._stream.read(1))[0]

    def read_uint16(self) -> int:
        """
        Read 2 bytes as an unsigned integer value from the stream.
        """
        return self._uint16.unpack(self._stream.read(2))[0]

    def read_uint16BE(self) -> int:
        """
        Read 2 bytes as an unsigned integer value from the stream.
        """
        return self._uint16BE.unpack(self._stream.read(2))[0]

    def read_int16(self) -> int:
        """
        Read 2 bytes as an unsigned integer value from the stream.
        """
        return self._int16.unpack(self._stream.read(2))[0]

    def read_int16BE(self) -> int:
        """
        Read 2 bytes as an unsigned integer value from the stream.
        """
        return self._int16BE.unpack(self._stream.read(2))[0]

    def read_uint32(self) -> int:
        """
        Read 4 bytes as an unsigned integer value from the stream.
        """
        return self._uint32.unpack(self._stream.read(4))[0]

    def read_uint32BE(self) -> int:
        """
        Read 4 bytes as an unsigned integer value from the stream.
        """
        return self._uint32BE.unpack(self._stream.read(4))[0]

    def read_int32(self) -> int:
        """
        Read 4 bytes as a signed integer value from the stream.
        """
        return self._int32.unpack(self._stream.read(4))[0]

    def read_int32BE(self) -> int:
        """
        Read 4 bytes as a signed integer value from the stream.
        """
        return self._int32BE.unpack(self._stream.read(4))[0]

    def read_uint64(self) -> int:
        """
        Read 8 bytes as an unsigned integer value from the stream.
        """
        return self._uint64.unpack(self._stream.read(8))[0]

    def read_uint64BE(self) -> int:
        """
        Read 8 bytes as an unsigned integer value from the stream.
        """
        return self._uint64BE.unpack(self._stream.read(8))[0]

    def read_int64(self) -> int:
        """
        Read 8 bytes as a signed integer value from the stream.
        """
        return self._int64.unpack(self._stream.read(8))[0]

    def read_int64BE(self) -> int:
        """
        Read 8 bytes as a signed integer value from the stream.
        """
        return self._int64BE.unpack(self._stream.read(8))[0]

    def read_var_int(self, max: int = sys.maxsize) -> int:
        """
        Read a integer that starts with a variable length indicator.

        The NEO network protocol supports encoded length indicating for saving space.
        See: :ref:`library-core-variable-length-encoding`

        See also:
            * :func:`~neo3.core.serialization.BinaryWriter.write_var_int`
            * :func:`~neo3.core.serialization.BinaryReader.read_var_string`
            * :func:`~neo3.core.serialization.BinaryReader.read_var_bytes`


        Args:
            max: (Optional) maximum number of bytes to read.

        Raises:
            ValueError: if the return value exceeds the `max` argument.
        """
        fb = int.from_bytes(self.read_byte(), 'little')
        if fb == 0:
            return fb

        if fb == 0xfd:
            value = self.read_uint16()
        elif fb == 0xfe:
            value = self.read_uint32()
        elif fb == 0xff:
            value = self.read_uint64()
        else:
            value = fb

        if value > max:
            raise ValueError("Invalid format")

        return value

    def read_var_bytes(self, max: int = sys.maxsize) -> bytes:
        """
        Read bytes that starts with a variable length indicator.

        The NEO network protocol supports encoded length indicating for saving space.
        See: :ref:`library-core-variable-length-encoding`

        See also:
            * :func:`~neo3.core.serialization.BinaryWriter.write_var_bytes`
            * :func:`~neo3.core.serialization.BinaryReader.read_var_int`
            * :func:`~neo3.core.serialization.BinaryReader.read_var_string`

        Note:
            This will try to read the amount of bytes as encoded in the stream but will read less if insufficient
            data is available.

        Args:
            max: (Optional) maximum number of bytes to read.
        """
        length = self.read_var_int(max)
        return self.read_bytes(length, _skip_length_check=True)

    def read_var_string(self, max: int = sys.maxsize) -> str:
        """
        Read a UTF-8 string that starts with a variable length indicator.

        The NEO network protocol supports encoded length indicating for saving space.
        See: :ref:`library-core-variable-length-encoding`

        See also:
            * :func:`~neo3.core.serialization.BinaryWriter.write_var_string`
            * :func:`~neo3.core.serialization.BinaryReader.read_var_int`
            * :func:`~neo3.core.serialization.BinaryReader.read_var_bytes`

        Args:
            max: (Optional) maximum number of bytes to read.

        Raises:
            ValueError: if decoding fails or insufficient data is present in the stream.
        """
        length = self.read_var_int(max)
        try:
            data = struct.unpack(f"{length}s", self._stream.read(length))[0]
            return data.decode('utf-8')
        except Exception as e:
            raise ValueError(str(e))

    def read_serializable(self, obj_type: Type[ISerializable_T]) -> ISerializable_T:
        """
        Read and deserialize an object of `obj_type` from the stream.

        Args:
            obj_type: the object class to deserialize into.
        """
        obj = obj_type._serializable_init()
        obj.deserialize(self)
        return obj

    def read_serializable_list(self, obj_type: Type[ISerializable_T], max: int = None) -> List[ISerializable_T]:
        """
        Read and deserialize a list of objects of `obj_type` from the stream.

        Expects to start with a `varint` list length indicator.

        Args:
            obj_type: the object class to deserialize into.
            max: read up to `max` objects from the stream.

        Returns:
            list[ISerializable]: list of deserialized objects.
        """
        obj_array = []
        count = self.read_var_int()
        if max and count > max:
            count = max

        try:
            for _ in range(count):
                obj = obj_type._serializable_init()
                obj.deserialize(self)
                obj_array.append(obj)
        except Exception as e:
            raise ValueError(f"Insufficient data - {str(e)}")
        return obj_array

    def close(self) -> None:
        """
        Close the internal stream to prevent resource leaking.

        Note:
            This is done automatically when using the context manager
        """
        if self._stream:
            self._stream.close()


class BinaryWriter(object):
    """
    A convenience class for writing data to byte streams.

    Context manager support is available to ensure proper cleanup of resources.

    Example:
    ::

        with serialization.BinaryWriter() as bw:
            bw.write_uint8(5)
            self.assertEqual(b'\\x05', bw._stream.getvalue())
    """

    def __init__(self, stream: Union[bytearray, bytes] = None) -> None:
        """
        Create an instance.

        Args:
            stream: a stream to operate on.
        """
        super(BinaryWriter, self).__init__()
        self._stream = BytesIO(stream) if stream else BytesIO()

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        self.close()

    def __len__(self):
        io = self._stream
        # Remember our current position
        cur_pos = io.tell()
        # Seek to the end of the File object
        io.seek(0, SEEK_END)
        # Remember position, which is equal to the full length
        full_size = io.tell()
        # Seek back to the current position
        io.seek(cur_pos)
        return full_size

    def write_bytes(self, value: bytes) -> int:
        """
        Write a `bytes` type to the stream.

        Args:
            value: array of bytes to write to the stream.

        Returns:
            int: the number of bytes written.
        """
        return self._stream.write(value)

    def _pack(self, fmt, data) -> int:
        """
        Write bytes by packing them according to the provided format `fmt`.

        For more information about the `fmt` format see: https://docs.python.org/3/library/struct.html

        Args:
            fmt (str): format string.
            data (object): the data to write to the raw stream.

        Returns:
            int: the number of bytes written.
        """
        return self.write_bytes(struct.pack(fmt, data))

    def write_bool(self, value: bool) -> int:
        """
        Pack the value as a bool and write 1 byte to the stream.

        Args:
            value: the boolean value to write.

        Returns:
            int: the number of bytes written.
        """
        return self._pack('?', value)

    def write_uint8(self, value) -> int:
        """

        Args:
            value: integer value to write to the stream.

        Returns:
            int: the number of bytes written.
        """
        return self.write_bytes(bytes([value]))

    def write_uint16(self, value: int, endian: str = "<") -> int:
        """
        Pack the value as an unsigned integer and write 2 bytes to the stream.

        Args:
            value: integer value to write to the stream.
            endian: specify the endianness. (Default) Little endian ('<'). Use '>' for big endian.

        Returns:
            int: the number of bytes written.
        """
        return self._pack('%sH' % endian, value)

    def write_uint32(self, value: int, endian: str = "<") -> int:
        """
        Pack the value as a signed integer and write 4 bytes to the stream.

        Args:
            value: integer value to write to the stream.
            endian: specify the endianness. (Default) Little endian ('<'). Use '>' for big endian.

        Returns:
            int: the number of bytes written.
        """
        return self._pack('%sI' % endian, value)

    def write_uint64(self, value: int, endian: str = "<") -> int:
        """
        Pack the value as an unsigned integer and write 8 bytes to the stream.

        Args:
            value: integer value to write to the stream.
            endian (str): specify the endianness. (Default) Little endian ('<'). Use '>' for big endian.

        Returns:
            int: the number of bytes written.
        """
        return self._pack('%sQ' % endian, value)

    def write_int16(self, value: int, endian: str = "<") -> int:
        """
        Pack the value as a signed integer and write 2 bytes to the stream.
        Args:
            value: integer value to write to the stream.
            endian: specify the endianness. (Default) Little endian ('<'). Use '>' for big endian.
        Returns:
            int: the number of bytes written.
        """
        return self._pack('%sh' % endian, value)

    def write_int32(self, value: int, endian: str = "<") -> int:
        """
        Pack the value as a signed integer and write 4 bytes to the stream.
        Args:
            value: integer value to write to the stream.
            endian: specify the endianness. (Default) Little endian ('<'). Use '>' for big endian.
        Returns:
            int: the number of bytes written.
        """
        return self._pack('%si' % endian, value)

    def write_int64(self, value: int, endian: str = "<") -> int:
        """
        Pack the value as an unsigned integer and write 8 bytes to the stream.
        Args:
            value: integer value to write to the stream.
            endian (str): specify the endianness. (Default) Little endian ('<'). Use '>' for big endian.
        Returns:
            int: the number of bytes written.
        """
        return self._pack('%sq' % endian, value)

    def write_var_string(self, value: str, encoding: str = "utf-8") -> int:
        """
        Write a string value to the stream.

        The NEO network protocol supports encoded length indicating for saving space.
        See: :ref:`library-core-variable-length-encoding`

        See also:
            * :func:`~neo3.core.serialization.BinaryReader.read_var_string`
            * :func:`~neo3.core.serialization.BinaryWriter.write_var_bytes`
            * :func:`~neo3.core.serialization.BinaryWriter.write_var_int`

        Args:
            value: string to write to the stream.
            encoding: string encoding format.
        """
        if type(value) is str:
            data = value.encode(encoding)

        length = len(data)
        self.write_var_int(length)
        written = self.write_bytes(data)
        return written

    def write_var_int(self, value: int, endian: str = "<") -> int:
        """
        Write an integer value in a space saving way to the stream.

        The NEO network protocol supports encoded length indicating for saving space.
        See: :ref:`library-core-variable-length-encoding`

        See also:
            * :func:`~neo3.core.serialization.BinaryReader.read_var_int`
            * :func:`~neo3.core.serialization.BinaryWriter.write_var_bytes`
            * :func:`~neo3.core.serialization.BinaryWriter.write_var_string`

        Args:
            value: integer value to write to the stream.
            endian: specify the endianness. (Default) Little endian ('<'). Use '>' for big endian.

        Raises:
            TypeError: if ``value`` is not of type int.
            ValueError: if `value` is < 0.

        Returns:
            int: the number of bytes written.
        """
        if not isinstance(value, int):
            raise TypeError('%s not int type.' % value)

        if value < 0:
            raise ValueError('%d too small.' % value)

        elif value < 0xfd:
            return self.write_bytes(bytes([value]))

        elif value <= 0xffff:
            self.write_bytes(bytes([0xfd]))
            return self.write_uint16(value, endian)

        elif value <= 0xFFFFFFFF:
            self.write_bytes(bytes([0xfe]))
            return self.write_uint32(value, endian)

        else:
            self.write_bytes(bytes([0xff]))
            return self.write_uint64(value, endian)

    def write_var_bytes(self, value: bytes, endian: str = "<") -> int:
        """
        Write bytes into a stream with variable length prefix.

        The NEO network protocol supports encoded length indicating for saving space. See
        :ref:`library-core-variable-length-encoding`

        See also:
            * :func:`~neo3.core.serialization.BinaryReader.read_var_bytes`
            * :func:`~neo3.core.serialization.BinaryWriter.write_var_int`
            * :func:`~neo3.core.serialization.BinaryWriter.write_var_string`

        Args:
            value: bytes to write to the stream.
            endian: specify the endianness. (Default) Little endian ('<'). Use '>' for big endian.

        Returns:
            int: the number of bytes written.
        """
        self.write_var_int(len(value), endian)
        return self.write_bytes(value)

    def write_serializable(self, obj_instance: ISerializable_T) -> None:
        """
        Serialize an object instance and write it to the stream.

        Args:
            obj_instance: the instance to serialize.
        """
        obj_instance.serialize(self)

    def write_serializable_list(self, objects: List[ISerializable_T]) -> None:
        """
        Serialize a list of objects and write them to the stream.

        Args:
            objects: a list of objects.
        """
        self.write_var_int(len(objects))
        for o in objects:
            o.serialize(self)

    def close(self) -> None:
        """
        Close the internal stream to prevent resource leaking.

        Note:
            This is done automatically when using the context manager
        """
        if self._stream:
            self._stream.close()

    def to_array(self) -> bytes:
        """
        Get the raw bytes from the underlying stream.
        """
        return self._stream.getvalue()



from __future__ import annotations
import binascii
from neo3.core import serialization
from typing import Type, Union

__all__ = ['UInt160', 'UInt256']


class _UIntBase(serialization.ISerializable):
    def __init__(self, num_bytes: int, data: Union[bytes, bytearray] = None) -> None:
        """

        Args:
            num_bytes:
            data:
        """
        super(_UIntBase, self).__init__()

        if data is None:
            self._data = bytes(num_bytes)
        else:
            if len(data) < num_bytes:
                raise ValueError(f"Invalid UInt: data length {len(data)} != specified num_bytes {num_bytes}")
            self._data = data[:num_bytes]

    def __len__(self) -> int:
        """ Count of data bytes. """
        return len(self._data)

    def __eq__(self, other) -> bool:
        if other is None:
            return False

        if not isinstance(other, _UIntBase):
            return False

        if other is self:
            return True

        if self._data == other._data:
            return True

        return False

    def __hash__(self):
        slice_length = 4 if len(self._data) >= 4 else len(self._data)
        return int.from_bytes(self._data[:slice_length], 'little')

    def __str__(self):
        """ Convert the data to a human readable format (data is in reverse byte order). """
        db = bytearray(self._data)
        db.reverse()
        return db.hex()

    def _compare_to(self, other) -> int:
        if not isinstance(other, _UIntBase):
            raise TypeError(f"Cannot compare {type(self).__name__} to type {type(other).__name__}")

        x = self._data
        y = other._data

        if len(x) != len(y):
            raise ValueError(f"Cannot compare {type(self).__name__} with length {len(x)} to {type(other).__name__} with"
                             f" length {len(y)}")

        length = len(x)

        for i in range(length - 1, 0, -1):
            if x[i] > y[i]:
                return 1
            if x[i] < y[i]:
                return -1

        return 0

    def __lt__(self, other):
        return self._compare_to(other) < 0

    def __gt__(self, other):
        return self._compare_to(other) > 0

    def __le__(self, other):
        return self._compare_to(other) <= 0

    def __ge__(self, other):
        return self._compare_to(other) >= 0

    def to_array(self) -> bytes:
        """
        Return an array of bytes representing the UInt

        Returns:
        """
        return bytes(self._data)

    @classmethod
    def _serializable_init(cls):
        return cls(b'\x00' * cls._BYTE_LEN)


class UInt160(_UIntBase):
    _BYTE_LEN = 20

    def __init__(self, data: bytes):
        """
        Initialize an instance.

        Args:
            data: hex escaped bytearray.
        """
        super(UInt160, self).__init__(num_bytes=self._BYTE_LEN, data=data)

    @classmethod
    def deserialize_from_bytes(cls: Type[UInt160], data: bytes) -> UInt160:
        """
        Parse data into an object instance.

        Args:
            data: hex escaped bytes.

        Raises:
            ValueError: if the length of the supplied bytearray is insufficient for the type.
        """
        if len(data) < cls._BYTE_LEN:
            raise ValueError(f"Insufficient data {len(data)} bytes is less than the required {cls._BYTE_LEN}")
        return cls(data[:cls._BYTE_LEN])

    @classmethod
    def from_string(cls: Type[UInt160], value: str) -> UInt160:
        """
        Try to parse a string into an instance.

        Note:
            NEO's string representation is in reverse byte order from the internal bytearray.


        Args:
            value: accepts the same input as :py:meth:`bytearray.fromhex`.

        Raises:
            ValueError: if the length of the supplied string does not match.
        """
        if len(value) != cls._BYTE_LEN * 2:
            raise ValueError(f"Invalid {cls.__name__} Format: {len(value)} chars != {cls._BYTE_LEN * 2} chars")
        reversed_data = bytearray.fromhex(value)
        reversed_data.reverse()
        return cls(data=reversed_data)

    @classmethod
    def zero(cls: Type[UInt160]) -> UInt160:
        """
        Returns:
            An instance initialized to zero.
        """
        return cls(data=bytes(20))

    def serialize(self, writer: serialization.BinaryWriter) -> None:
        """
        Serialize the object into a binary stream.

        Args:
            writer: instance.
        """
        writer.write_bytes(self._data)

    def deserialize(self, reader: serialization.BinaryReader) -> None:
        """
        Deserialize the object from a binary stream.

        Args:
            reader: instance.
        """
        self._data = bytes(reader.read_bytes(self._BYTE_LEN))


class UInt256(_UIntBase):
    _BYTE_LEN = 32

    def __init__(self, data: bytes):
        """
        Initialize an instance.

        Args:
            data: hex escaped bytearray.
        """
        super(UInt256, self).__init__(num_bytes=self._BYTE_LEN, data=data)

    @classmethod
    def deserialize_from_bytes(cls: Type[UInt256], data: bytes) -> UInt256:
        """
        Parse data into an object instance.

        Args:
            data: hex escaped bytes.

        Raises:
            ValueError: if the length of the supplied bytearray is insufficient for the type.
        """
        if len(data) < cls._BYTE_LEN:
            raise ValueError(f"Insufficient data {len(data)} bytes is less than the required {cls._BYTE_LEN}")
        return cls(data[:cls._BYTE_LEN])

    @classmethod
    def from_string(cls: Type[UInt256], value: str) -> UInt256:
        """
        Try to parse a string into an instance.

        Note:
            NEO's string representation is in reverse byte order from the internal bytearray.


        Args:
            value: accepts the same input as :py:meth:`bytearray.fromhex`.

        Raises:
            ValueError: if the length of the supplied string does not match.
        """
        if len(value) != cls._BYTE_LEN * 2:
            raise ValueError(f"Invalid {cls.__name__} Format: {len(value)} chars != {cls._BYTE_LEN * 2} chars")
        reversed_data = bytearray.fromhex(value)
        reversed_data.reverse()
        return cls(data=reversed_data)

    @classmethod
    def zero(cls: Type[UInt256]) -> UInt256:
        """
        Returns:
            An instance initialized to zero.
        """

        return cls(data=bytes(cls._BYTE_LEN))

    def serialize(self, writer: serialization.BinaryWriter) -> None:
        """
        Serialize the object into a binary stream.

        Args:
            writer: instance.
        """
        writer.write_bytes(self._data)

    def deserialize(self, reader: serialization.BinaryReader) -> None:
        """
        Deserialize the object from a binary stream.

        Args:
            reader: instance.
        """
        self._data = bytes(reader.read_bytes(self._BYTE_LEN))



from datetime import datetime
from typing import List


class NodeWeight:
    SPEED_RECORD_COUNT = 3
    SPEED_INIT_VALUE = 100 * 1024 ^ 2  # Start with a big speed of 100 MB/s

    REQUEST_TIME_RECORD_COUNT = 3

    def __init__(self, node_id: int):
        """
        An internal class for tracking and calculating node weight data.

        Is used by ``NodeManager`` for selecting the next node candidate.

        Args:
            node_id: the :attr:`~neo3.network.node.NeoNode.id` of the node this weight belongs to.
        """

        #: The :attr:`~neo3.network.node.NeoNode.id` of the node this weight belongs to.
        self.id: int = node_id

        #: in number of bytes per second
        self.speed: List[float] = [self.SPEED_INIT_VALUE] * self.SPEED_RECORD_COUNT
        self.timeout_count = 0
        self.error_response_count = 0
        now = datetime.utcnow().timestamp() * 1000  # milliseconds
        self.request_time: List[float] = [now] * self.REQUEST_TIME_RECORD_COUNT

    def append_new_speed(self, speed: float) -> None:
        """
        Add a speed value to the collection from which the average speed is calculated.

        The average is calculated based on 3 values. The last value is popped from the collection.

        Args:
            speed: in number of bytes per second
        """
        # remove oldest
        self.speed.pop(-1)
        # add new
        self.speed.insert(0, speed)

    def append_new_request_time(self) -> None:
        """
        Add a request time value to the collection from which the average request time is calculated

        The average is calculated based on 3 values. The last value is popped from the collection.
        """

        self.request_time.pop(-1)

        now = datetime.utcnow().timestamp() * 1000  # milliseconds
        self.request_time.insert(0, now)

    def _avg_speed(self) -> float:
        return sum(self.speed) / self.SPEED_RECORD_COUNT

    def _avg_request_time(self) -> float:
        avg_request_time: float = 0
        now = datetime.utcnow().timestamp() * 1000  # milliseconds

        for t in self.request_time:
            avg_request_time += now - t

        avg_request_time = avg_request_time / self.REQUEST_TIME_RECORD_COUNT
        return avg_request_time

    def weight(self) -> float:
        """
        A score indicating the quality of the node. Higher is better.

        Nodes with the highest speed and the longest time between querying for data have the highest weight
        and will be accessed first by the syncmanager unless their error/timeout count is higher. This distributes the
        load across nodes.
        """
        weight = self._avg_speed() + self._avg_request_time()

        # punish errors and timeouts harder than slower speeds and more recent access
        if self.error_response_count:
            weight /= self.error_response_count + 1  # make sure we at least always divide by 2

        if self.timeout_count:
            weight /= self.timeout_count + 1
        return weight

    def __lt__(self, other):
        return self.weight() < other.weight()

    def __repr__(self):
        return f"<{self.__class__.__name__} at {hex(id(self))}> weight:{self.weight():.2f}"



import hashlib
from neo3.core import types
from typing import List, Optional


class _MerkleTreeNode:
    def __init__(self, hash: types.UInt256 = None):
        self.hash = hash if hash else types.UInt256.zero()  # type: types.UInt256
        self.parent = None  # type: Optional[_MerkleTreeNode]
        self.left_child = None  # type: Optional[_MerkleTreeNode]
        self.right_child = None  # type: Optional[_MerkleTreeNode]

    def is_leaf(self) -> bool:
        """
        Return `True` if the node is a leaf.
        """
        return self.left_child is None and self.right_child is None

    def is_root(self) -> bool:
        """
        Return `True` if the node is the root node.
        """
        return self.parent is None


class MerkleTree:
    def __init__(self, hashes: List[types.UInt256]):
        """

        Args:
            hashes: the list of hashes to build the tree from.

        Raises:
            ValueError: if the `hashes` list is empty.
        """
        if len(hashes) == 0:
            self.has_root = False
            return
        else:
            self.has_root = True

        self.root = self._build(leaves=[_MerkleTreeNode(h) for h in hashes])
        _depth = 1
        i = self.root
        while i.left_child is not None:
            _depth += 1
            i = i.left_child
        self.depth = _depth

    def to_hash_array(self) -> List[types.UInt256]:
        """
        Create a list of hashes the Merkle tree is build up from.

        Note: does not include the Merkle root hash.
        """
        hashes: List[types.UInt256] = []
        if self.has_root:
            MerkleTree._depth_first_search(self.root, hashes)
        return hashes

    @staticmethod
    def _depth_first_search(node: _MerkleTreeNode, hashes: List[types.UInt256]) -> None:
        # if left is None then Right is also always None, but it helps the static type checker understand this
        # otherwise it thinks it might go to the else branch and the second call is then invalid
        if node.left_child is None or node.right_child is None:
            hashes.append(node.hash)
        else:
            MerkleTree._depth_first_search(node.left_child, hashes)
            MerkleTree._depth_first_search(node.right_child, hashes)

    @staticmethod
    def _build(leaves: List[_MerkleTreeNode]) -> _MerkleTreeNode:
        if len(leaves) == 0:
            raise ValueError('Leaves must have length')
        if len(leaves) == 1:
            return leaves[0]

        num_parents = int((len(leaves) + 1) / 2)
        parents = [_MerkleTreeNode() for i in range(0, num_parents)]

        for i in range(0, num_parents):
            node = parents[i]
            node.left_child = leaves[i * 2]
            leaves[i * 2].parent = node
            if (i * 2 + 1 == len(leaves)):
                node.right_child = node.left_child
            else:
                node.right_child = leaves[i * 2 + 1]
                leaves[i * 2 + 1].parent = node

            data = node.left_child.hash.to_array() + node.right_child.hash.to_array()
            hashed_data = hashlib.sha256(hashlib.sha256(data).digest()).digest()
            node.hash = types.UInt256(data=hashed_data)

        return MerkleTree._build(parents)

    @staticmethod
    def compute_root(hashes: List[types.UInt256]) -> types.UInt256:
        """
        Compute the Merkle root hash from a list of hashes.

        Args:
            hashes:

        Raises:
             ValueError: if the `hashes` list is empty.
        """
        if len(hashes) == 0:
            return types.UInt256.zero()
        if len(hashes) == 1:
            return hashes[0]
        tree = MerkleTree(hashes)
        return tree.root.hash



import unittest
import binascii
import lz4
from unittest.mock import patch, call
from neo3.network import payloads, message
from neo3.core.types.uint import UInt256
from neo3.core import serialization

class NetworkMessageTestCase(unittest.TestCase):
    def test_create_no_payload(self):
        m = message.Message(message.MessageType.PING, payload=None)
        self.assertEqual(message.MessageType.PING, m.type)
        self.assertEqual(message.MessageConfig.NONE, m.config)

    def test_create_inv_message(self):
        hashes = [UInt256.zero()]
        inv_payload = payloads.InventoryPayload(payloads.InventoryType.BLOCK, hashes)
        m = message.Message(message.MessageType.INV, inv_payload)
        data = m.to_array()

        self.assertEqual(message.MessageType.INV, m.type)
        self.assertEqual(message.MessageConfig.NONE, m.config)
        self.assertIsInstance(m.payload, payloads.InventoryPayload)

        """
            Taken from constructing the same object in C#
            
            UInt256[] hashes = { UInt256.Zero };
            var inv_payload = InvPayload.Create(InventoryType.Block, hashes);
            ISerializable message = Message.Create(MessageCommand.Inv, inv_payload);

            using (MemoryStream ms = new MemoryStream())
            using (BinaryWriter writer = new BinaryWriter(ms))
            {
                message.Serialize(writer);
                writer.Flush();
                byte[] data = ms.ToArray();
                Console.WriteLine($"b\'{BitConverter.ToString(data).Replace("-","")}\'");
            }          

        """
        expected_data = binascii.unhexlify(b'0027222C010000000000000000000000000000000000000000000000000000000000000000')
        self.assertEqual(expected_data, data)

    def test_create_compressed_inv_message(self):
        hashes = [UInt256.zero(), UInt256.zero(), UInt256.zero(), UInt256.zero()]
        inv_payload = payloads.InventoryPayload(payloads.InventoryType.BLOCK, hashes)
        m = message.Message(message.MessageType.INV, inv_payload)
        data = m.to_array() # triggers payload compression

        self.assertEqual(message.MessageType.INV, m.type)
        self.assertEqual(message.MessageConfig.COMPRESSED, m.config)
        self.assertIsInstance(m.payload, payloads.InventoryPayload)

        """
        Data created in the same fashion as how it's done in test_create_inv_message()
        The deviation is `hashes` now contains 4 x UInt256.zero()
        """

        expected_data = binascii.unhexlify(b'012711820000003F2C0400010067500000000000')
        self.assertEqual(expected_data, data)

    def test_inv_message_deserialization(self):
        # see test_create_compressed_inv_message() how it was obtained
        raw_data = binascii.unhexlify(b'012711820000003F2C0400010067500000000000')
        m = message.Message.deserialize_from_bytes(raw_data)
        self.assertIsInstance(m.payload, payloads.InventoryPayload)
        self.assertEqual(132, len(m))

    def test_deserialization_with_not_enough_data(self):
        with self.assertRaises(ValueError) as context:
            m = message.Message.deserialize_from_bytes(bytearray(2))
        self.assertEqual(str(context.exception), "Could not read byte from empty stream")

    def test_deserialization_without_payload(self):
        # some message types like PING/PONG have no payload
        m = message.Message(message.MessageType.PING)
        data = m.to_array()
        m2 = message.Message.deserialize_from_bytes(data)
        self.assertEqual(message.MessageType.PING, m2.type)
        self.assertEqual(0, len(m2.payload))

    def test_deserialization_from_stream(self):
        # see test_create_compressed_inv_message() how it was obtained
        raw_data = binascii.unhexlify(b'012711820000003F2C0400010067500000000000')
        with serialization.BinaryReader(raw_data) as br:
            m = message.Message(message.MessageType.DEFAULT)
            m.deserialize(br)
            self.assertEqual(m.type, message.MessageType.INV)
            self.assertEqual(m.payload.type, payloads.inventory.InventoryType.BLOCK)

    def test_deserialization_with_unsupported_payload_type(self):
        hashes = [UInt256.zero()]
        inv_payload = payloads.InventoryPayload(payloads.InventoryType.BLOCK, hashes)
        m = message.Message(message.MessageType.ALERT, inv_payload)

        m2 = message.Message.deserialize_from_bytes(m.to_array())
        self.assertIsInstance(m2.payload, payloads.EmptyPayload)

    def test_deserialization_erroneous_compressed_data(self):
        # see test_create_compressed_inv_message() how it was obtained
        raw_data = binascii.unhexlify(b'01270D3F020400010067500000000000')

        with patch('lz4.block.decompress') as lz4_mock:
            with self.assertRaises(ValueError) as context:
                lz4_mock.side_effect = lz4.block.LZ4BlockError()
                m = message.Message.deserialize_from_bytes(raw_data)
            self.assertEqual("Invalid payload data - decompress failed", str(context.exception))

    def test_payload_from_data(self):
        with patch('neo3.core.serialization.BinaryReader') as br:
            reader = br.return_value.__enter__.return_value
            message.Message._payload_from_data(message.MessageType.INV, b'')
            message.Message._payload_from_data(message.MessageType.GETBLOCKBYINDEX, b'')
            message.Message._payload_from_data(message.MessageType.VERSION, b'')
            message.Message._payload_from_data(message.MessageType.VERACK, b'')
            message.Message._payload_from_data(message.MessageType.BLOCK, b'')
            message.Message._payload_from_data(message.MessageType.HEADERS, b'')
            message.Message._payload_from_data(message.MessageType.PING, b'')
            message.Message._payload_from_data(message.MessageType.PONG, b'')
            message.Message._payload_from_data(message.MessageType.ADDR, b'')
            message.Message._payload_from_data(message.MessageType.TRANSACTION, b'')

            calls = [
                call(payloads.InventoryPayload),
                call(payloads.GetBlockByIndexPayload),
                call(payloads.VersionPayload),
                call(payloads.EmptyPayload),
                call(payloads.Block),
                call(payloads.HeadersPayload),
                call(payloads.PingPayload),
                call(payloads.PingPayload),
                call(payloads.AddrPayload),
                call(payloads.Transaction)
            ]
            reader.read_serializable.assert_has_calls(calls, any_order=False)



import unittest
from neo3 import contracts
from neo3.core import types


class ContractParameterDefinitionTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.expected = {'name': "Main", 'type': 'String'}

    def test_to_json(self):
        cpd = contracts.ContractParameterDefinition("Main", contracts.ContractParameterType.STRING)
        self.assertEqual(self.expected, cpd.to_json())

    def test_from_json(self):
        cpd = contracts.ContractParameterDefinition.from_json(self.expected)
        self.assertEqual("Main", cpd.name)
        self.assertEqual(contracts.ContractParameterType.STRING, cpd.type)

        with self.assertRaises(KeyError) as context:
            json_without_name = {'type': contracts.ContractParameterType.ANY}
            contracts.ContractParameterDefinition.from_json(json_without_name)
        self.assertIn('name', str(context.exception))

        with self.assertRaises(KeyError) as context:
            json_without_type = {'name': "Main"}
            contracts.ContractParameterDefinition.from_json(json_without_type)
        self.assertIn('type', str(context.exception))

    def test_eq(self):
        cpd = contracts.ContractParameterDefinition.from_json(self.expected)
        cpd2 = contracts.ContractParameterDefinition.from_json(self.expected)
        self.assertFalse(cpd == object())
        self.assertTrue(cpd == cpd2)


class ContractEventDescriptorTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.expected = {"name": "MainEvent", "parameters": [{"name": "param", "type": "String"}]}
        cls.parameters = [contracts.ContractParameterDefinition("param", contracts.ContractParameterType.STRING)]

    def test_to_json(self):
        ced = contracts.ContractEventDescriptor("MainEvent", self.parameters)
        self.assertEqual(self.expected, ced.to_json())

    def test_from_json(self):
        ced = contracts.ContractEventDescriptor.from_json(self.expected)
        self.assertEqual("MainEvent", ced.name)
        self.assertEqual(self.parameters, ced.parameters)

        with self.assertRaises(KeyError) as context:
            json_without_name = self.expected.copy()
            json_without_name.pop('name')
            contracts.ContractEventDescriptor.from_json(json_without_name)
        self.assertIn('name', str(context.exception))

        with self.assertRaises(KeyError) as context:
            json_without_param = self.expected.copy()
            json_without_param.pop('parameters')
            contracts.ContractEventDescriptor.from_json(json_without_param)
        self.assertIn('parameters', str(context.exception))

    def test_eq(self):
        ced = contracts.ContractEventDescriptor.from_json(self.expected)
        ced2 = contracts.ContractEventDescriptor.from_json(self.expected)
        self.assertFalse(ced == object())
        self.assertTrue(ced == ced2)


class ContractMethodDescriptorTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.parameters = [contracts.ContractParameterDefinition("param", contracts.ContractParameterType.STRING)]
        cls.expected = {"name": "MainMethod",
                        "offset": 0,
                        "parameters": [cls.parameters[0].to_json()],
                        "returntype": "Boolean",
                        "safe": True}

    def test_to_json(self):
        cmd = contracts.ContractMethodDescriptor("MainMethod",
                                                 0,
                                                 self.parameters, contracts.ContractParameterType.BOOLEAN,
                                                 True
                                                 )
        self.assertEqual(self.expected, cmd.to_json())

    def test_from_json(self):
        cmd = contracts.ContractMethodDescriptor.from_json(self.expected)
        self.assertEqual("MainMethod", cmd.name)
        self.assertEqual(self.parameters, cmd.parameters)
        self.assertEqual(0, cmd.offset)
        self.assertEqual(contracts.ContractParameterType.BOOLEAN, cmd.return_type)
        self.assertTrue(cmd.safe)

        with self.assertRaises(KeyError) as context:
            json_without_return_type = self.expected.copy()
            json_without_return_type.pop('returntype')
            contracts.ContractMethodDescriptor.from_json(json_without_return_type)
        self.assertIn('returntype', str(context.exception))

    def test_eq(self):
        cmd = contracts.ContractMethodDescriptor.from_json(self.expected)
        cmd2 = contracts.ContractMethodDescriptor.from_json(self.expected)
        self.assertFalse(cmd == object())
        self.assertTrue(cmd == cmd2)


class AbiTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        """
        var methods = new ContractMethodDescriptor[]
        {
            new ContractMethodDescriptor {
                Name = "main_entry",
                Offset = 0,
                Parameters = new ContractParameterDefinition[0],
                ReturnType = ContractParameterType.Integer,
                Safe = true
            }
        };
        var events = new ContractEventDescriptor[]
        {
            new ContractEventDescriptor() { Name = "main_event", Parameters = new ContractParameterDefinition[0]}
        };
        var abi = new ContractAbi()
        {
            Methods = methods,
            Events = events
        };
        Console.WriteLine(abi.ToJson());
        """
        cls.method1 = contracts.ContractMethodDescriptor(
            name="main_entry",
            offset=0,
            parameters=[],
            return_type=contracts.ContractParameterType.INTEGER,
            safe=True
        )
        cls.methods = [cls.method1]
        cls.event = contracts.ContractEventDescriptor(
            name="main_event",
            parameters=[]
        )
        cls.events = [cls.event]
        # captured from C#
        cls.expected_json = {"methods":[{"name":"main_entry","parameters":[],"returntype":"Integer","offset":0,"safe":True}],"events":[{"name":"main_event","parameters":[]}]}

    def test_to_json(self):
        abi = contracts.ContractABI(
            methods=self.methods,
            events=self.events
        )

        self.assertEqual(self.expected_json, abi.to_json())

    def test_from_json(self):
        # if "test_to_json" passes, then we know we can use our defined class
        # attributes to validate the from_json() results
        abi = contracts.ContractABI.from_json(self.expected_json)
        self.assertEqual(self.methods, abi.methods)
        self.assertEqual(self.events, abi.events)

    def test_eq(self):
        abi = contracts.ContractABI.from_json(self.expected_json)
        abi2 = contracts.ContractABI.from_json(self.expected_json)
        self.assertFalse(abi == object())
        self.assertTrue(abi == abi2)

    def test_get_method(self):
        abi = contracts.ContractABI.from_json(self.expected_json)
        self.assertIsNone(abi.get_method("bad_method", 0))
        method = abi.get_method("main_entry", 0)
        self.assertIsInstance(method, contracts.ContractMethodDescriptor)
        self.assertEqual("main_entry", method.name)



from __future__ import annotations
import abc
from typing import cast
from neo3 import vm, contracts
from neo3.contracts.interop import register


class IIterator(abc.ABC):
    @abc.abstractmethod
    def next(self) -> bool:
        """ Advance the iterator """

    @abc.abstractmethod
    def value(self) -> vm.StackItem:
        """ Get the current value """


class StorageIterator(IIterator):
    def __init__(self,
                 generator,
                 prefix_length: int,
                 options: contracts.FindOptions,
                 reference_counter: vm.ReferenceCounter):
        self.it = generator
        self.prefix_len = prefix_length
        self.options = options
        self.reference_counter = reference_counter
        self._pair = None

    def next(self) -> bool:
        try:
            self._pair = next(self.it)
            return True
        except StopIteration:
            self._pair = None
            return False

    def value(self) -> vm.StackItem:
        if self._pair is None:
            raise ValueError("Cannot call 'value' without having advanced the iterator at least once")
        key = self._pair[0].key
        value = self._pair[1].value
        if contracts.FindOptions.REMOVE_PREFIX in self.options:
            key = key[self.prefix_len:]

        if contracts.FindOptions.DESERIALIZE_VALUES in self.options:
            item: vm.StackItem = contracts.BinarySerializer.deserialize(value, 1024, self.reference_counter)
        else:
            item = vm.ByteStringStackItem(value)

        if contracts.FindOptions.PICK_FIELD0 in self.options:
            item = cast(vm.ArrayStackItem, item)
            item = item[0]
        elif contracts.FindOptions.PICK_FIELD1 in self.options:
            item = cast(vm.ArrayStackItem, item)
            item = item[1]

        if contracts.FindOptions.KEYS_ONLY in self.options:
            return vm.ByteStringStackItem(key)

        if contracts.FindOptions.VALUES_ONLY in self.options:
            return vm.ByteStringStackItem(value)

        return vm.StructStackItem(self.reference_counter, [vm.ByteStringStackItem(key), item])


@register("System.Iterator.Next", 1 << 15, contracts.CallFlags.NONE)
def iterator_next(engine: contracts.ApplicationEngine, iterator: IIterator) -> bool:
    return iterator.next()


@register("System.Iterator.Value", 1 << 4, contracts.CallFlags.NONE)
def iterator_value(engine: contracts.ApplicationEngine, iterator: IIterator) -> vm.StackItem:
    return iterator.value()



import os


def setup(app):
    current_dir = os.path.abspath(os.path.dirname(__file__))
    app.add_html_theme(
        'neo3', current_dir)

    return {
        'parallel_read_safe': True,
        'parallel_write_safe': True,
    }


from __future__ import annotations
from typing import Iterator, Tuple, Dict, List, Optional, TYPE_CHECKING
from neo3 import storage
from neo3.core import types
from neo3.network import payloads
from contextlib import suppress
from copy import deepcopy

if TYPE_CHECKING:
    from neo3 import contracts


class MemoryDB(storage.IDBImplementation):
    BLOCK = 'blocks'
    BLOCK_HEIGHT_MAP = 'blocksmap'
    BLOCK_BEST_HEIGHT = 'blockheight'
    CONTRACT = 'contracts'
    CONTRACT_ID = 'contract_id'
    STORAGE = 'storages'
    TX = 'transactions'

    def __init__(self, options: dict = None):
        self.db: Dict[str, dict] = {
            self.BLOCK: {},
            self.BLOCK_HEIGHT_MAP: {},
            self.CONTRACT: {},
            self.STORAGE: {},
            self.TX: {}
        }
        self._best_block_height = -1
        self._contract_id = 0

    def get_snapshotview(self) -> MemorySnapshot:
        return MemorySnapshot(self)

    def _internal_bestblockheight_get(self):
        if self._best_block_height == -1:
            raise KeyError
        return self._best_block_height

    def _internal_bestblockheight_put(self, height: int, batch=None):
        if batch:
            batch.put(self.BLOCK_BEST_HEIGHT, height, height)
        else:
            self._best_block_height = height

    def _internal_bestblockheight_update(self, height: int, batch=None):
        self._internal_bestblockheight_put(height, batch)

    def _internal_block_put(self, block: payloads.Block, batch: WriteBatch = None) -> None:
        if batch:
            batch.put(self.BLOCK, block.hash(), block)
        else:
            self.db[self.BLOCK][block.hash()] = block
            self.db[self.BLOCK_HEIGHT_MAP][block.index] = block.hash()

        stored_value = -1
        with suppress(KeyError):
            stored_value = self._internal_bestblockheight_get()

        if block.index > stored_value:
            self._best_block_height = block.index

    def _internal_block_update(self, block: payloads.Block, batch: WriteBatch = None) -> None:
        self._internal_block_put(block, batch)

    def _internal_block_delete(self, hash: types.UInt256, batch: WriteBatch = None) -> None:
        if batch:
            batch.delete(self.BLOCK, hash)
        else:
            with suppress(KeyError):
                block = self.db[self.BLOCK].pop(hash)
                self.db[self.BLOCK_HEIGHT_MAP].pop(block.index)

    def _internal_block_get(self, hash: types.UInt256) -> payloads.Block:
        value = self.db[self.BLOCK].get(hash, None)
        if value is None:
            raise KeyError
        return deepcopy(value)

    def _internal_block_get_by_height(self, height: int) -> payloads.Block:
        block_hash = self.db[self.BLOCK_HEIGHT_MAP].get(height, None)
        if block_hash is None:
            raise KeyError

        return self._internal_block_get(block_hash)

    def _internal_block_all(self) -> Iterator[payloads.Block]:
        for block in self.db[self.BLOCK].values():
            yield deepcopy(block)

    def _internal_contractid_get(self):
        if self._contract_id == -1:
            raise KeyError
        return self._contract_id

    def _internal_contractid_put(self, new_id: int, batch=None):
        if batch:
            batch.put(self.CONTRACT_ID, new_id, new_id)
        else:
            self._contract_id = new_id

    def _internal_contractid_update(self, new_id: int, batch=None):
        self._internal_contractid_update(new_id, batch)

    def _internal_contract_put(self, contract: contracts.ContractState, batch: WriteBatch = None) -> None:
        if batch:
            batch.put(self.CONTRACT, contract.hash, contract)
        else:
            self.db[self.CONTRACT][contract.hash] = contract

    def _internal_contract_update(self, contract: contracts.ContractState, batch: WriteBatch = None) -> None:
        self._internal_contract_put(contract, batch)

    def _internal_contract_delete(self, hash: types.UInt160, batch: WriteBatch = None) -> None:
        if batch:
            batch.delete(self.CONTRACT, hash)
        else:
            with suppress(KeyError):
                self.db[self.CONTRACT].pop(hash)

    def _internal_contract_get(self, contract_hash: types.UInt160) -> contracts.ContractState:
        value = self.db[self.CONTRACT].get(contract_hash, None)
        if value is None:
            raise KeyError

        return deepcopy(value)

    def _internal_contract_all(self) -> Iterator[contracts.ContractState]:
        for contract in self.db[self.CONTRACT].values():
            yield deepcopy(contract)

    def _internal_storage_put(self, key: storage.StorageKey,
                              value: storage.StorageItem,
                              batch: WriteBatch = None) -> None:
        if batch:
            batch.put(self.STORAGE, key, value)
        else:
            self.db[self.STORAGE][key] = value

    def _internal_storage_update(self, key: storage.StorageKey,
                                 value: storage.StorageItem,
                                 batch: WriteBatch = None) -> None:
        self._internal_storage_put(key, value, batch)

    def _internal_storage_delete(self, key: storage.StorageKey, batch: WriteBatch = None) -> None:
        if batch:
            batch.delete(self.STORAGE, key)
        else:
            with suppress(KeyError):
                self.db[self.STORAGE].pop(key)

    def _internal_storage_get(self, key: storage.StorageKey) -> storage.StorageItem:
        value = self.db[self.STORAGE].get(key, None)
        if value is None:
            raise KeyError

        return deepcopy(value)

    def _internal_storage_all(self, contract_id: Optional[int] = None) -> Iterator[Tuple[storage.StorageKey,
                                                                                         storage.StorageItem]]:
        for k, v in self.db[self.STORAGE].items():
            if contract_id:
                if contract_id == k.id:
                    yield deepcopy(k), deepcopy(v)
            else:
                yield deepcopy(k), deepcopy(v)

    def _internal_storage_find(self, key_prefix: bytes) -> Iterator[Tuple[storage.StorageKey, storage.StorageItem]]:
        script_hash_len = 20
        for k, v in self.db[self.STORAGE].items():
            # k is of type StorageKey, which starts with a 4-byte contract id.
            # We skip this and search only in the `key` attribute
            if k.to_array().startswith(key_prefix):
                yield deepcopy(k), deepcopy(v)

    def _internal_storage_seek(self,
                               key_prefix: bytes,
                               seek_direction="forward") -> Iterator[Tuple[storage.StorageKey, storage.StorageItem]]:
        if seek_direction == "forward":
            view = iter(self.db[self.STORAGE].items())
        else:
            view = reversed(self.db[self.STORAGE].items())
        for key, value in view:  # type: storage.StorageKey, storage.StorageItem
            if key.to_array().startswith(key_prefix):
                yield deepcopy(key), deepcopy(value)

    def _internal_transaction_put(self, transaction: payloads.Transaction, batch: WriteBatch = None) -> None:
        if batch:
            batch.put(self.TX, transaction.hash(), transaction)
        else:
            self.db[self.TX][transaction.hash()] = transaction

    def _internal_transaction_update(self, transaction: payloads.Transaction, batch: WriteBatch = None) -> None:
        self._internal_transaction_put(transaction, batch)

    def _internal_transaction_delete(self, hash: types.UInt256, batch: WriteBatch = None) -> None:
        if batch:
            batch.delete(self.TX, hash)
        else:
            with suppress(KeyError):
                self.db[self.TX].pop(hash)

    def _internal_transaction_get(self, hash: types.UInt256) -> payloads.Transaction:
        value = self.db[self.TX].get(hash, None)
        if value is None:
            raise KeyError

        return deepcopy(value)

    def _internal_transaction_all(self) -> Iterator[payloads.Transaction]:
        for tx in self.db[self.TX].values():
            yield deepcopy(tx)

    def write_batch(self, batch) -> None:
        for table, action, pair in batch:
            if action == 'delete':
                item = self.db[table].pop(pair[0])
                if table == self.BLOCK:
                    self.db[self.BLOCK_HEIGHT_MAP].pop(item.index)
            elif action in ['update', 'add']:
                if table == self.BLOCK_BEST_HEIGHT:
                    self._best_block_height = pair[0]
                    continue
                self.db[table][pair[0]] = pair[1]
                if table == self.BLOCK:
                    # pair = (UInt256, Block)
                    self.db[self.BLOCK_HEIGHT_MAP][pair[1].index] = pair[0]


class WriteBatch:
    def __init__(self):
        self.statements = []

    def __iter__(self):
        for s in self.statements:
            yield s

    def put(self, table, key, value) -> None:
        self.statements.append((table, 'add', (key, value)))

    def delete(self, table, key) -> None:
        self.statements.append((table, 'delete', (key, None)))


class MemorySnapshot(storage.Snapshot):
    def __init__(self, db: MemoryDB):
        super(MemorySnapshot, self).__init__()
        self._db = db
        self._batch = WriteBatch()

        self._block_cache = MemoryDBCachedBlockAccess(db, self._batch)
        self._storage_cache = MemoryDBCachedStorageAccess(db, self._batch)
        self._contract_cache = MemoryDBCachedContractAccess(db, self._batch)
        self._tx_cache = MemoryDBCachedTXAccess(db, self._batch)
        self._block_height_cache = MemoryBestBlockHeightAttribute(db, self._batch)

    def commit(self) -> None:
        super(MemorySnapshot, self).commit()
        self._db.write_batch(self._batch)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        # nothing to do
        pass


class MemoryBestBlockHeightAttribute(storage.AttributeCache):
    def __init__(self, db, batch):
        super(MemoryBestBlockHeightAttribute, self).__init__()
        self._db = db
        self._batch = batch

    def _get_internal(self):
        return self._db._internal_bestblockheight_get()

    def _update_internal(self, value):
        self._db._internal_bestblockheight_update(value, self._batch)


class MemoryDBCachedBlockAccess(storage.CachedBlockAccess):
    def __init__(self, db, batch):
        super(MemoryDBCachedBlockAccess, self).__init__(db)
        self._batch = batch

    def commit(self) -> None:
        keys_to_delete: List[types.UInt256] = []
        for trackable in self._dictionary.values():  # trackable.item: payloads.Block
            if trackable.state == storage.TrackState.ADDED:
                self._db._internal_block_put(trackable.item, self._batch)
                trackable.state = storage.TrackState.NONE
            elif trackable.state == storage.TrackState.CHANGED:
                self._db._internal_block_update(trackable.item, self._batch)
                trackable.state = storage.TrackState.NONE
            elif trackable.state == storage.TrackState.DELETED:
                self._db._internal_block_delete(trackable.item.hash(), self._batch)
                keys_to_delete.append(trackable.key)
        for key in keys_to_delete:
            with suppress(KeyError):
                self._dictionary.pop(key)
        self._changeset.clear()

    def create_snapshot(self):
        return storage.CloneBlockCache(self._db, self)


class MemoryDBCachedContractAccess(storage.CachedContractAccess):
    def __init__(self, db, batch):
        super(MemoryDBCachedContractAccess, self).__init__(db)
        self._batch = batch

    def commit(self) -> None:
        keys_to_delete: List[types.UInt160] = []
        for trackable in self.get_changeset():  # trackable.item: contracts.ContractState
            if trackable.state == storage.TrackState.ADDED:
                self._db._internal_contract_put(trackable.item, self._batch)
                trackable.state = storage.TrackState.NONE
            elif trackable.state == storage.TrackState.CHANGED:
                self._db._internal_contract_update(trackable.item, self._batch)
                trackable.state = storage.TrackState.NONE
            elif trackable.state == storage.TrackState.DELETED:
                self._db._internal_contract_delete(trackable.item.hash, self._batch)
                keys_to_delete.append(trackable.key)
        for key in keys_to_delete:
            with suppress(KeyError):
                self._dictionary.pop(key)
        self._changeset.clear()

    def create_snapshot(self):
        return storage.CloneContractCache(self._db, self)


class MemoryDBCachedStorageAccess(storage.CachedStorageAccess):
    def __init__(self, db, batch):
        super(MemoryDBCachedStorageAccess, self).__init__(db)
        self._batch = batch

    def commit(self) -> None:
        keys_to_delete: List[storage.StorageKey] = []
        for trackable in self.get_changeset():
            if trackable.state == storage.TrackState.ADDED:
                self._db._internal_storage_put(trackable.key, trackable.item, self._batch)
                trackable.state = storage.TrackState.NONE
            elif trackable.state == storage.TrackState.CHANGED:
                self._db._internal_storage_update(trackable.key, trackable.item, self._batch)
                trackable.state = storage.TrackState.NONE
            elif trackable.state == storage.TrackState.DELETED:
                self._db._internal_storage_delete(trackable.key, self._batch)
                keys_to_delete.append(trackable.key)
        for key in keys_to_delete:
            with suppress(KeyError):
                self._dictionary.pop(key)
        self._changeset.clear()

    def create_snapshot(self):
        return storage.CloneStorageCache(self._db, self)


class MemoryDBCachedTXAccess(storage.CachedTXAccess):
    def __init__(self, db, batch):
        super(MemoryDBCachedTXAccess, self).__init__(db)
        self._batch = batch

    def commit(self) -> None:
        keys_to_delete: List[types.UInt256] = []
        for trackable in self._dictionary.values():  # trackable.item: payloads.Transaction
            if trackable.state == storage.TrackState.ADDED:
                self._db._internal_transaction_put(trackable.item, self._batch)
                trackable.state = storage.TrackState.NONE
            elif trackable.state == storage.TrackState.CHANGED:
                self._db._internal_transaction_update(trackable.item, self._batch)
                trackable.state = storage.TrackState.NONE
            elif trackable.state == storage.TrackState.DELETED:
                self._db._internal_transaction_delete(trackable.item.hash(), self._batch)
                keys_to_delete.append(trackable.key)
        for key in keys_to_delete:
            with suppress(KeyError):
                self._dictionary.pop(key)
        self._changeset.clear()

    def create_snapshot(self):
        return storage.CloneTXCache(self._db, self)
